'''
Extensions to ConsIndShockModel concerning models with preference shocks.
It currently only two models:

1) An extension of ConsIndShock, but with an iid lognormal multiplicative shock each period.
2) A combination of (1) and ConsKinkedR, demonstrating how to construct a new model
   by inheriting from multiple classes.
'''

import numpy as np
from HARK.utilities import approxMeanOneLognormal
from ConsIndShockModel import IndShockConsumerType, ConsumerSolution, ConsIndShockSolver, \
                                   ValueFunc, MargValueFunc, KinkedRconsumerType, ConsKinkedRsolver
from HARK.interpolation import LinearInterpOnInterp1D, LinearInterp, CubicInterp, LowerEnvelope

class PrefShockConsumerType(IndShockConsumerType):
    '''
    A class for representing consumers who experience multiplicative shocks to
    utility each period, specified as iid lognormal.
    '''
    shock_vars_ = IndShockConsumerType.shock_vars_ + ['PrefShkNow']

    def __init__(self,cycles=1,time_flow=True,**kwds):
        '''
        Instantiate a new ConsumerType with given data, and construct objects
        to be used during solution (income distribution, assets grid, etc).
        See ConsumerParameters.init_pref_shock for a dictionary of
        the keywords that should be passed to the constructor.

        Parameters
        ----------
        cycles : int
            Number of times the sequence of periods should be solved.
        time_flow : boolean
            Whether time is currently "flowing" forward for this instance.

        Returns
        -------
        None
        '''
        IndShockConsumerType.__init__(self,cycles=cycles,time_flow=time_flow,**kwds)
        self.solveOnePeriod = solveConsPrefShock # Choose correct solver

    def update(self):
        '''
        Updates the assets grid, income process, terminal period solution, and
        preference shock process.  A very slight extension of IndShockConsumerType.update()
        for the preference shock model.

        Parameters
        ----------
        None

        Returns
        -------
        None
        '''
        IndShockConsumerType.update(self)  # Update assets grid, income process, terminal solution
        self.updatePrefShockProcess()     # Update the discrete preference shock process

    def updatePrefShockProcess(self):
        '''
        Make a discrete preference shock structure for each period in the cycle
        for this agent type, storing them as attributes of self for use in the
        solution (and other methods).

        Parameters
        ----------
        none

        Returns
        -------
        none
        '''
        time_orig = self.time_flow
        self.timeFwd()

        PrefShkDstn = [] # discrete distributions of preference shocks
        for t in range(len(self.PrefShkStd)):
            PrefShkStd = self.PrefShkStd[t]
            PrefShkDstn.append(approxMeanOneLognormal(N=self.PrefShkCount,
                                                      sigma=PrefShkStd,tail_N=self.PrefShk_tail_N))

        # Store the preference shocks in self (time-varying) and restore time flow
        self.PrefShkDstn = PrefShkDstn
        self.addToTimeVary('PrefShkDstn')
        if not time_orig:
            self.timeRev()

    def getShocks(self):
        '''
        Gets permanent and transitory income shocks for this period as well as preference shocks.

        Parameters
        ----------
        None

        Returns
        -------
        None
        '''
        IndShockConsumerType.getShocks(self) # Get permanent and transitory income shocks
        PrefShkNow = np.zeros(self.AgentCount) # Initialize shock array
        for t in range(self.T_cycle):
            these = t == self.t_cycle
            N = np.sum(these)
            if N > 0:
                PrefShkNow[these] = self.RNG.permutation(approxMeanOneLognormal(N,sigma=self.PrefShkStd[t])[1])
        self.PrefShkNow = PrefShkNow

    def getControls(self):
        '''
        Calculates consumption for each consumer of this type using the consumption functions.

        Parameters
        ----------
        None

        Returns
        -------
        None
        '''
        cNrmNow = np.zeros(self.AgentCount) + np.nan
        for t in range(self.T_cycle):
            these = t == self.t_cycle
            cNrmNow[these] = self.solution[t].cFunc(self.mNrmNow[these],self.PrefShkNow[these])
        self.cNrmNow = cNrmNow
        return None


    def calcBoundingValues(self):
        '''
        Calculate human wealth plus minimum and maximum MPC in an infinite
        horizon model with only one period repeated indefinitely.  Store results
        as attributes of self.  Human wealth is the present discounted value of
        expected future income after receiving income this period, ignoring mort-
        ality.  The maximum MPC is the limit of the MPC as m --> mNrmMin.  The
        minimum MPC is the limit of the MPC as m --> infty.

        NOT YET IMPLEMENTED FOR THIS CLASS

        Parameters
        ----------
        None

        Returns
        -------
        None
        '''
        raise NotImplementedError()

    def makeEulerErrorFunc(self,mMax=100,approx_inc_dstn=True):
        '''
        Creates a "normalized Euler error" function for this instance, mapping
        from market resources to "consumption error per dollar of consumption."
        Stores result in attribute eulerErrorFunc as an interpolated function.
        Has option to use approximate income distribution stored in self.IncomeDstn
        or to use a (temporary) very dense approximation.

        NOT YET IMPLEMENTED FOR THIS CLASS

        Parameters
        ----------
        mMax : float
            Maximum normalized market resources for the Euler error function.
        approx_inc_dstn : Boolean
            Indicator for whether to use the approximate discrete income distri-
            bution stored in self.IncomeDstn[0], or to use a very accurate
            discrete approximation instead.  When True, uses approximation in
            IncomeDstn; when False, makes and uses a very dense approximation.

        Returns
        -------
        None
        '''
        raise NotImplementedError()


class KinkyPrefConsumerType(PrefShockConsumerType,KinkedRconsumerType):
    '''
    A class for representing consumers who experience multiplicative shocks to
    utility each period, specified as iid lognormal and different interest rates
    on borrowing vs saving.
    '''
    def __init__(self,cycles=1,time_flow=True,**kwds):
        '''
        Instantiate a new ConsumerType with given data, and construct objects
        to be used during solution (income distribution, assets grid, etc).
        See ConsumerParameters.init_kinky_pref for a dictionary of the keywords
        that should be passed to the constructor.

        Parameters
        ----------
        cycles : int
            Number of times the sequence of periods should be solved.
        time_flow : boolean
            Whether time is currently "flowing" forward for this instance.

        Returns
        -------
        None
        '''
        IndShockConsumerType.__init__(self,**kwds)
        self.solveOnePeriod = solveConsKinkyPref # Choose correct solver
        self.addToTimeInv('Rboro','Rsave')
        self.delFromTimeInv('Rfree')

    def getRfree(self): # Specify which getRfree to use
        return KinkedRconsumerType.getRfree(self)

###############################################################################

class ConsPrefShockSolver(ConsIndShockSolver):
    '''
    A class for solving the one period consumption-saving problem with risky
    income (permanent and transitory shocks) and multiplicative shocks to utility
    each period.
    '''
    def __init__(self,solution_next,IncomeDstn,PrefShkDstn,LivPrb,DiscFac,CRRA,
                      Rfree,PermGroFac,BoroCnstArt,aXtraGrid,vFuncBool,CubicBool):
        '''
        Constructor for a new solver for problems with risky income, a different
        interest rate on borrowing and saving, and multiplicative shocks to utility.

        Parameters
        ----------
        solution_next : ConsumerSolution
            The solution to the succeeding one period problem.
        IncomeDstn : [np.array]
            A list containing three arrays of floats, representing a discrete
            approximation to the income process between the period being solved
            and the one immediately following (in solution_next). Order: event
            probabilities, permanent shocks, transitory shocks.
        PrefShkDstn : [np.array]
            Discrete distribution of the multiplicative utility shifter.  Order:
            probabilities, preference shocks.
        LivPrb : float
            Survival probability; likelihood of being alive at the beginning of
            the succeeding period.
        DiscFac : float
            Intertemporal discount factor for future utility.
        CRRA : float
            Coefficient of relative risk aversion.
        Rfree : float
            Risk free interest factor on end-of-period assets.
        PermGroGac : float
            Expected permanent income growth factor at the end of this period.
        BoroCnstArt: float or None
            Borrowing constraint for the minimum allowable assets to end the
            period with.  If it is less than the natural borrowing constraint,
            then it is irrelevant; BoroCnstArt=None indicates no artificial bor-
            rowing constraint.
        aXtraGrid: np.array
            Array of "extra" end-of-period asset values-- assets above the
            absolute minimum acceptable level.
        vFuncBool: boolean
            An indicator for whether the value function should be computed and
            included in the reported solution.
        CubicBool: boolean
            An indicator for whether the solver should use cubic or linear inter-
            polation.

        Returns
        -------
        None
        '''
        ConsIndShockSolver.__init__(self,solution_next,IncomeDstn,LivPrb,DiscFac,CRRA,
                      Rfree,PermGroFac,BoroCnstArt,aXtraGrid,vFuncBool,CubicBool)
        self.PrefShkPrbs = PrefShkDstn[0]
        self.PrefShkVals = PrefShkDstn[1]

    def getPointsForInterpolation(self,EndOfPrdvP,aNrmNow):
        '''
        Find endogenous interpolation points for each asset point and each
        discrete preference shock.

        Parameters
        ----------
        EndOfPrdvP : np.array
            Array of end-of-period marginal values.
        aNrmNow : np.array
            Array of end-of-period asset values that yield the marginal values
            in EndOfPrdvP.

        Returns
        -------
        c_for_interpolation : np.array
            Consumption points for interpolation.
        m_for_interpolation : np.array
            Corresponding market resource points for interpolation.
        '''
        c_base       = self.uPinv(EndOfPrdvP)
        PrefShkCount = self.PrefShkVals.size
        PrefShk_temp = np.tile(np.reshape(self.PrefShkVals**(1.0/self.CRRA),(PrefShkCount,1)),
                               (1,c_base.size))
        self.cNrmNow = np.tile(c_base,(PrefShkCount,1))*PrefShk_temp
        self.mNrmNow = self.cNrmNow + np.tile(aNrmNow,(PrefShkCount,1))

        # Add the bottom point to the c and m arrays
        m_for_interpolation = np.concatenate((self.BoroCnstNat*np.ones((PrefShkCount,1)),
                                              self.mNrmNow),axis=1)
        c_for_interpolation = np.concatenate((np.zeros((PrefShkCount,1)),self.cNrmNow),axis=1)
        return c_for_interpolation,m_for_interpolation

    def usePointsForInterpolation(self,cNrm,mNrm,interpolator):
        '''
        Make a basic solution object with a consumption function and marginal
        value function (unconditional on the preference shock).

        Parameters
        ----------
        cNrm : np.array
            Consumption points for interpolation.
        mNrm : np.array
            Corresponding market resource points for interpolation.
        interpolator : function
            A function that constructs and returns a consumption function.

        Returns
        -------
        solution_now : ConsumerSolution
            The solution to this period's consumption-saving problem, with a
            consumption function, marginal value function, and minimum m.
        '''
        # Make the preference-shock specific consumption functions
        PrefShkCount = self.PrefShkVals.size
        cFunc_list   = []
        for j in range(PrefShkCount):
            MPCmin_j         = self.MPCminNow*self.PrefShkVals[j]**(1.0/self.CRRA)
            cFunc_this_shock = LowerEnvelope(LinearInterp(mNrm[j,:],cNrm[j,:],
                                             intercept_limit=self.hNrmNow*MPCmin_j,
                                             slope_limit=MPCmin_j),self.cFuncNowCnst)
            cFunc_list.append(cFunc_this_shock)

        # Combine the list of consumption functions into a single interpolation
        cFuncNow = LinearInterpOnInterp1D(cFunc_list,self.PrefShkVals)

        # Make the ex ante marginal value function (before the preference shock)
        m_grid = self.aXtraGrid + self.mNrmMinNow
        vP_vec = np.zeros_like(m_grid)
        for j in range(PrefShkCount): # numeric integration over the preference shock
            vP_vec += self.uP(cFunc_list[j](m_grid))*self.PrefShkPrbs[j]*self.PrefShkVals[j]
        vPnvrs_vec = self.uPinv(vP_vec)
        vPfuncNow  = MargValueFunc(LinearInterp(m_grid,vPnvrs_vec),self.CRRA)

        # Store the results in a solution object and return it
        solution_now = ConsumerSolution(cFunc=cFuncNow, vPfunc=vPfuncNow, mNrmMin=self.mNrmMinNow)
        return solution_now

    def makevFunc(self,solution):
        '''
        Make the beginning-of-period value function (unconditional on the shock).

        Parameters
        ----------
        solution : ConsumerSolution
            The solution to this single period problem, which must include the
            consumption function.

        Returns
        -------
        vFuncNow : ValueFunc
            A representation of the value function for this period, defined over
            normalized market resources m: v = vFuncNow(m).
        '''
        # Compute expected value and marginal value on a grid of market resources,
        # accounting for all of the discrete preference shocks
        PrefShkCount = self.PrefShkVals.size
        mNrm_temp   = self.mNrmMinNow + self.aXtraGrid
        vNrmNow     = np.zeros_like(mNrm_temp)
        vPnow       = np.zeros_like(mNrm_temp)
        for j in range(PrefShkCount):
            this_shock  = self.PrefShkVals[j]
            this_prob   = self.PrefShkPrbs[j]
            cNrmNow     = solution.cFunc(mNrm_temp,this_shock*np.ones_like(mNrm_temp))
            aNrmNow     = mNrm_temp - cNrmNow
            vNrmNow    += this_prob*(this_shock*self.u(cNrmNow) + self.EndOfPrdvFunc(aNrmNow))
            vPnow      += this_prob*this_shock*self.uP(cNrmNow)

        # Construct the beginning-of-period value function
        vNvrs        = self.uinv(vNrmNow) # value transformed through inverse utility
        vNvrsP       = vPnow*self.uinvP(vNrmNow)
        mNrm_temp    = np.insert(mNrm_temp,0,self.mNrmMinNow)
        vNvrs        = np.insert(vNvrs,0,0.0)
        vNvrsP       = np.insert(vNvrsP,0,self.MPCmaxEff**(-self.CRRA/(1.0-self.CRRA)))
        MPCminNvrs   = self.MPCminNow**(-self.CRRA/(1.0-self.CRRA))
        vNvrsFuncNow = CubicInterp(mNrm_temp,vNvrs,vNvrsP,MPCminNvrs*self.hNrmNow,MPCminNvrs)
        vFuncNow     = ValueFunc(vNvrsFuncNow,self.CRRA)
        return vFuncNow


def solveConsPrefShock(solution_next,IncomeDstn,PrefShkDstn,
                       LivPrb,DiscFac,CRRA,Rfree,PermGroFac,BoroCnstArt,
                       aXtraGrid,vFuncBool,CubicBool):
    '''
    Solves a single period of a consumption-saving model with preference shocks
    to marginal utility.  Problem is solved using the method of endogenous gridpoints.

    Parameters
    ----------
    solution_next : ConsumerSolution
        The solution to the succeeding one period problem.
    IncomeDstn : [np.array]
        A list containing three arrays of floats, representing a discrete
        approximation to the income process between the period being solved
        and the one immediately following (in solution_next). Order: event
        probabilities, permanent shocks, transitory shocks.
    PrefShkDstn : [np.array]
        Discrete distribution of the multiplicative utility shifter.  Order:
        probabilities, preference shocks.
    LivPrb : float
        Survival probability; likelihood of being alive at the beginning of
        the succeeding period.
    DiscFac : float
        Intertemporal discount factor for future utility.
    CRRA : float
        Coefficient of relative risk aversion.
    Rfree : float
        Risk free interest factor on end-of-period assets.
    PermGroGac : float
        Expected permanent income growth factor at the end of this period.
    BoroCnstArt: float or None
        Borrowing constraint for the minimum allowable assets to end the
        period with.  If it is less than the natural borrowing constraint,
        then it is irrelevant; BoroCnstArt=None indicates no artificial bor-
        rowing constraint.
    aXtraGrid: np.array
        Array of "extra" end-of-period asset values-- assets above the
        absolute minimum acceptable level.
    vFuncBool: boolean
        An indicator for whether the value function should be computed and
        included in the reported solution.
    CubicBool: boolean
        An indicator for whether the solver should use cubic or linear inter-
        polation.

    Returns
    -------
    solution: ConsumerSolution
        The solution to the single period consumption-saving problem.  Includes
        a consumption function cFunc (using linear splines), a marginal value
        function vPfunc, a minimum acceptable level of normalized market re-
        sources mNrmMin, normalized human wealth hNrm, and bounding MPCs MPCmin
        and MPCmax.  It might also have a value function vFunc.  The consumption
        function is defined over normalized market resources and the preference
        shock, c = cFunc(m,PrefShk), but the (marginal) value function is defined
        unconditionally on the shock, just before it is revealed.
    '''
    solver = ConsPrefShockSolver(solution_next,IncomeDstn,PrefShkDstn,LivPrb,
                             DiscFac,CRRA,Rfree,PermGroFac,BoroCnstArt,aXtraGrid,
                             vFuncBool,CubicBool)
    solver.prepareToSolve()
    solution = solver.solve()
    return solution

###############################################################################

class ConsKinkyPrefSolver(ConsPrefShockSolver,ConsKinkedRsolver):
    '''
    A class for solving the one period consumption-saving problem with risky
    income (permanent and transitory shocks), multiplicative shocks to utility
    each period, and a different interest rate on saving vs borrowing.
    '''
    def __init__(self,solution_next,IncomeDstn,PrefShkDstn,LivPrb,DiscFac,CRRA,
                      Rboro,Rsave,PermGroFac,BoroCnstArt,aXtraGrid,vFuncBool,CubicBool):
        '''
        Constructor for a new solver for problems with risky income, a different
        interest rate on borrowing and saving, and multiplicative shocks to utility.

        Parameters
        ----------
        solution_next : ConsumerSolution
            The solution to the succeeding one period problem.
        IncomeDstn : [np.array]
            A list containing three arrays of floats, representing a discrete
            approximation to the income process between the period being solved
            and the one immediately following (in solution_next). Order: event
            probabilities, permanent shocks, transitory shocks.
        PrefShkDstn : [np.array]
            Discrete distribution of the multiplicative utility shifter.  Order:
            probabilities, preference shocks.
        LivPrb : float
            Survival probability; likelihood of being alive at the beginning of
            the succeeding period.
        DiscFac : float
            Intertemporal discount factor for future utility.
        CRRA : float
            Coefficient of relative risk aversion.
        Rboro: float
            Interest factor on assets between this period and the succeeding
            period when assets are negative.
        Rsave: float
            Interest factor on assets between this period and the succeeding
            period when assets are positive.
        PermGroGac : float
            Expected permanent income growth factor at the end of this period.
        BoroCnstArt: float or None
            Borrowing constraint for the minimum allowable assets to end the
            period with.  If it is less than the natural borrowing constraint,
            then it is irrelevant; BoroCnstArt=None indicates no artificial bor-
            rowing constraint.
        aXtraGrid: np.array
            Array of "extra" end-of-period asset values-- assets above the
            absolute minimum acceptable level.
        vFuncBool: boolean
            An indicator for whether the value function should be computed and
            included in the reported solution.
        CubicBool: boolean
            An indicator for whether the solver should use cubic or linear inter-
            polation.

        Returns
        -------
        None
        '''
        ConsKinkedRsolver.__init__(self,solution_next,IncomeDstn,LivPrb,DiscFac,CRRA,
                      Rboro,Rsave,PermGroFac,BoroCnstArt,aXtraGrid,vFuncBool,CubicBool)
        self.PrefShkPrbs = PrefShkDstn[0]
        self.PrefShkVals = PrefShkDstn[1]


def solveConsKinkyPref(solution_next,IncomeDstn,PrefShkDstn,
                       LivPrb,DiscFac,CRRA,Rboro,Rsave,PermGroFac,BoroCnstArt,
                       aXtraGrid,vFuncBool,CubicBool):
    '''
    Solves a single period of a consumption-saving model with preference shocks
    to marginal utility and a different interest rate on saving vs borrowing.
    Problem is solved using the method of endogenous gridpoints.

    Parameters
    ----------
    solution_next : ConsumerSolution
        The solution to the succeeding one period problem.
    IncomeDstn : [np.array]
        A list containing three arrays of floats, representing a discrete
        approximation to the income process between the period being solved
        and the one immediately following (in solution_next). Order: event
        probabilities, permanent shocks, transitory shocks.
    PrefShkDstn : [np.array]
        Discrete distribution of the multiplicative utility shifter.  Order:
        probabilities, preference shocks.
    LivPrb : float
        Survival probability; likelihood of being alive at the beginning of
        the succeeding period.
    DiscFac : float
        Intertemporal discount factor for future utility.
    CRRA : float
        Coefficient of relative risk aversion.
    Rboro: float
        Interest factor on assets between this period and the succeeding
        period when assets are negative.
    Rsave: float
        Interest factor on assets between this period and the succeeding
        period when assets are positive.
    PermGroGac : float
        Expected permanent income growth factor at the end of this period.
    BoroCnstArt: float or None
        Borrowing constraint for the minimum allowable assets to end the
        period with.  If it is less than the natural borrowing constraint,
        then it is irrelevant; BoroCnstArt=None indicates no artificial bor-
        rowing constraint.
    aXtraGrid: np.array
        Array of "extra" end-of-period asset values-- assets above the
        absolute minimum acceptable level.
    vFuncBool: boolean
        An indicator for whether the value function should be computed and
        included in the reported solution.
    CubicBool: boolean
        An indicator for whether the solver should use cubic or linear inter-
        polation.

    Returns
    -------
    solution: ConsumerSolution
        The solution to the single period consumption-saving problem.  Includes
        a consumption function cFunc (using linear splines), a marginal value
        function vPfunc, a minimum acceptable level of normalized market re-
        sources mNrmMin, normalized human wealth hNrm, and bounding MPCs MPCmin
        and MPCmax.  It might also have a value function vFunc.  The consumption
        function is defined over normalized market resources and the preference
        shock, c = cFunc(m,PrefShk), but the (marginal) value function is defined
        unconditionally on the shock, just before it is revealed.
    '''
    solver = ConsKinkyPrefSolver(solution_next,IncomeDstn,PrefShkDstn,LivPrb,
                             DiscFac,CRRA,Rboro,Rsave,PermGroFac,BoroCnstArt,
                             aXtraGrid,vFuncBool,CubicBool)
    solver.prepareToSolve()
    solution = solver.solve()
    return solution

###############################################################################

if __name__ == '__main__':
    import ConsumerParameters as Params
    import matplotlib.pyplot as plt
    from HARK.utilities import plotFuncs
    from time import clock
    mystr = lambda number : "{:.4f}".format(number)

    do_simulation = True

    # Make and solve a preference shock consumer
    PrefShockExample = PrefShockConsumerType(**Params.init_preference_shocks)
    PrefShockExample.cycles = 0 # Infinite horizon

    t_start = clock()
    PrefShockExample.solve()
    t_end = clock()
    print('Solving a preference shock consumer took ' + str(t_end-t_start) + ' seconds.')

    # Plot the consumption function at each discrete shock
    m = np.linspace(PrefShockExample.solution[0].mNrmMin,5,200)
    print('Consumption functions at each discrete shock:')
    for j in range(PrefShockExample.PrefShkDstn[0][1].size):
        PrefShk = PrefShockExample.PrefShkDstn[0][1][j]
        c = PrefShockExample.solution[0].cFunc(m,PrefShk*np.ones_like(m))
        plt.plot(m,c)
    plt.show()

    print('Consumption function (and MPC) when shock=1:')
    c = PrefShockExample.solution[0].cFunc(m,np.ones_like(m))
    k = PrefShockExample.solution[0].cFunc.derivativeX(m,np.ones_like(m))
    plt.plot(m,c)
    plt.plot(m,k)
    plt.show()

    if PrefShockExample.vFuncBool:
        print('Value function (unconditional on shock):')
        plotFuncs(PrefShockExample.solution[0].vFunc,PrefShockExample.solution[0].mNrmMin+0.5,5)

    # Test the simulator for the pref shock class
    if do_simulation:
        PrefShockExample.T_sim = 120
        PrefShockExample.track_vars = ['cNrmNow']
        PrefShockExample.makeShockHistory() # This is optional
        PrefShockExample.initializeSim()
        PrefShockExample.simulate()

    ###########################################################################

    # Make and solve a "kinky preferece" consumer, whose model combines KinkedR and PrefShock
    KinkyPrefExample = KinkyPrefConsumerType(**Params.init_kinky_pref)
    KinkyPrefExample.cycles = 0 # Infinite horizon

    t_start = clock()
    KinkyPrefExample.solve()
    t_end = clock()
    print('Solving a kinky preference consumer took ' + str(t_end-t_start) + ' seconds.')

    # Plot the consumption function at each discrete shock
    m = np.linspace(KinkyPrefExample.solution[0].mNrmMin,5,200)
    print('Consumption functions at each discrete shock:')
    for j in range(KinkyPrefExample.PrefShkDstn[0][1].size):
        PrefShk = KinkyPrefExample.PrefShkDstn[0][1][j]
        c = KinkyPrefExample.solution[0].cFunc(m,PrefShk*np.ones_like(m))
        plt.plot(m,c)
    plt.show()

    print('Consumption function (and MPC) when shock=1:')
    c = KinkyPrefExample.solution[0].cFunc(m,np.ones_like(m))
    k = KinkyPrefExample.solution[0].cFunc.derivativeX(m,np.ones_like(m))
    plt.plot(m,c)
    plt.plot(m,k)
    plt.show()

    if KinkyPrefExample.vFuncBool:
        print('Value function (unconditional on shock):')
        plotFuncs(KinkyPrefExample.solution[0].vFunc,KinkyPrefExample.solution[0].mNrmMin+0.5,5)

    # Test the simulator for the kinky preference class
    if do_simulation:
        KinkyPrefExample.T_sim = 120
        KinkyPrefExample.track_vars = ['cNrmNow','PrefShkNow']
        KinkyPrefExample.initializeSim()
        KinkyPrefExample.simulate()



'''
Consumption-saving models that also include medical spending.
'''

import numpy as np
from scipy.optimize import brentq
from HARK import HARKobject
from HARK.utilities import approxLognormal, addDiscreteOutcomeConstantMean, CRRAutilityP_inv,\
                           CRRAutility, CRRAutility_inv, CRRAutility_invP, CRRAutilityPP,\
                           makeGridExpMult, NullFunc
from HARK.simulation import drawLognormal
from ConsIndShockModel import ConsumerSolution
from HARK.interpolation import BilinearInterpOnInterp1D, TrilinearInterp, BilinearInterp, CubicInterp,\
                               LinearInterp, LowerEnvelope3D, UpperEnvelope, LinearInterpOnInterp1D,\
                               VariableLowerBoundFunc3D
from ConsGenIncProcessModel import ConsGenIncProcessSolver, PersistentShockConsumerType,\
                                     ValueFunc2D, MargValueFunc2D, MargMargValueFunc2D, \
                                     VariableLowerBoundFunc2D
from copy import deepcopy

utility_inv   = CRRAutility_inv
utilityP_inv  = CRRAutilityP_inv
utility       = CRRAutility
utility_invP  = CRRAutility_invP
utilityPP     = CRRAutilityPP


class MedShockPolicyFunc(HARKobject):
    '''
    Class for representing the policy function in the medical shocks model: opt-
    imal consumption and medical care for given market resources, permanent income,
    and medical need shock.  Always obeys Con + MedPrice*Med = optimal spending.
    '''
    distance_criteria = ['xFunc','cFunc','MedPrice']

    def __init__(self,xFunc,xLvlGrid,MedShkGrid,MedPrice,CRRAcon,CRRAmed,xLvlCubicBool=False,
                 MedShkCubicBool=False):
        '''
        Make a new MedShockPolicyFunc.

        Parameters
        ----------
        xFunc : function
            Optimal total spending as a function of market resources, permanent
            income, and the medical need shock.
        xLvlGrid : np.array
            1D array of total expenditure levels.
        MedShkGrid : np.array
            1D array of medical shocks.
        MedPrice : float
            Relative price of a unit of medical care.
        CRRAcon : float
            Coefficient of relative risk aversion for consumption.
        CRRAmed : float
            Coefficient of relative risk aversion for medical care.
        xLvlCubicBool : boolean
            Indicator for whether cubic spline interpolation (rather than linear)
            should be used in the xLvl dimension.
        MedShkCubicBool : boolean
            Indicator for whether bicubic interpolation should be used; only
            operative when xLvlCubicBool=True.

        Returns
        -------
        None
        '''
        # Store some of the inputs in self
        self.MedPrice = MedPrice
        self.xFunc = xFunc

        # Calculate optimal consumption at each combination of mLvl and MedShk.
        cLvlGrid = np.zeros((xLvlGrid.size,MedShkGrid.size)) # Initialize consumption grid
        for i in range(xLvlGrid.size):
            xLvl = xLvlGrid[i]
            for j in range(MedShkGrid.size):
                MedShk = MedShkGrid[j]
                if xLvl == 0: # Zero consumption when mLvl = 0
                    cLvl = 0.0
                elif MedShk == 0: # All consumption when MedShk = 0
                    cLvl = xLvl
                else:
                    optMedZeroFunc = lambda c : (MedShk/MedPrice)**(-1.0/CRRAcon)*\
                                     ((xLvl-c)/MedPrice)**(CRRAmed/CRRAcon) - c
                    cLvl = brentq(optMedZeroFunc,0.0,xLvl) # Find solution to FOC
                cLvlGrid[i,j] = cLvl

        # Construct the consumption function and medical care function
        if xLvlCubicBool:
            if MedShkCubicBool:
                raise NotImplementedError(), 'Bicubic interpolation not yet implemented'
            else:
                xLvlGrid_tiled   = np.tile(np.reshape(xLvlGrid,(xLvlGrid.size,1)),
                                           (1,MedShkGrid.size))
                MedShkGrid_tiled = np.tile(np.reshape(MedShkGrid,(1,MedShkGrid.size)),
                                           (xLvlGrid.size,1))
                dfdx = (CRRAmed/(CRRAcon*MedPrice))*(MedShkGrid_tiled/MedPrice)**(-1.0/CRRAcon)*\
                       ((xLvlGrid_tiled - cLvlGrid)/MedPrice)**(CRRAmed/CRRAcon - 1.0)
                dcdx = dfdx/(dfdx + 1.0)
                dcdx[0,:] = dcdx[1,:] # approximation; function goes crazy otherwise
                dcdx[:,0] = 1.0 # no Med when MedShk=0, so all x is c
                cFromxFunc_by_MedShk = []
                for j in range(MedShkGrid.size):
                    cFromxFunc_by_MedShk.append(CubicInterp(xLvlGrid,cLvlGrid[:,j],dcdx[:,j]))
                cFunc = LinearInterpOnInterp1D(cFromxFunc_by_MedShk,MedShkGrid)
        else:
            cFunc = BilinearInterp(cLvlGrid,xLvlGrid,MedShkGrid)
        self.cFunc = cFunc

    def __call__(self,mLvl,pLvl,MedShk):
        '''
        Evaluate optimal consumption and medical care at given levels of market
        resources, permanent income, and medical need shocks.

        Parameters
        ----------
        mLvl : np.array
            Market resource levels.
        pLvl : np.array
            Permanent income levels; should be same size as mLvl.
        MedShk : np.array
            Medical need shocks; should be same size as mLvl.

        Returns
        -------
        cLvl : np.array
            Optimal consumption for each point in (xLvl,MedShk).
        Med : np.array
            Optimal medical care for each point in (xLvl,MedShk).
        '''
        xLvl = self.xFunc(mLvl,pLvl,MedShk)
        cLvl = self.cFunc(xLvl,MedShk)
        Med  = (xLvl-cLvl)/self.MedPrice
        return cLvl,Med

    def derivativeX(self,mLvl,pLvl,MedShk):
        '''
        Evaluate the derivative of consumption and medical care with respect to
        market resources at given levels of market resources, permanent income,
        and medical need shocks.

        Parameters
        ----------
        mLvl : np.array
            Market resource levels.
        pLvl : np.array
            Permanent income levels; should be same size as mLvl.
        MedShk : np.array
            Medical need shocks; should be same size as mLvl.

        Returns
        -------
        dcdm : np.array
            Derivative of consumption with respect to market resources for each
            point in (xLvl,MedShk).
        dMeddm : np.array
            Derivative of medical care with respect to market resources for each
            point in (xLvl,MedShk).
        '''
        xLvl = self.xFunc(mLvl,pLvl,MedShk)
        dxdm = self.xFunc.derivativeX(mLvl,pLvl,MedShk)
        dcdx = self.cFunc.derivativeX(xLvl,MedShk)
        dcdm = dxdm*dcdx
        dMeddm = (dxdm - dcdm)/self.MedPrice
        return dcdm,dMeddm

    def derivativeY(self,mLvl,pLvl,MedShk):
        '''
        Evaluate the derivative of consumption and medical care with respect to
        permanent income at given levels of market resources, permanent income,
        and medical need shocks.

        Parameters
        ----------
        mLvl : np.array
            Market resource levels.
        pLvl : np.array
            Permanent income levels; should be same size as mLvl.
        MedShk : np.array
            Medical need shocks; should be same size as mLvl.

        Returns
        -------
        dcdp : np.array
            Derivative of consumption with respect to permanent income for each
            point in (xLvl,MedShk).
        dMeddp : np.array
            Derivative of medical care with respect to permanent income for each
            point in (xLvl,MedShk).
        '''
        xLvl = self.xFunc(mLvl,pLvl,MedShk)
        dxdp = self.xFunc.derivativeY(mLvl,pLvl,MedShk)
        dcdx = self.cFunc.derivativeX(xLvl,MedShk)
        dcdp = dxdp*dcdx
        dMeddp = (dxdp - dcdp)/self.MedPrice
        return dcdp,dMeddp

    def derivativeZ(self,mLvl,pLvl,MedShk):
        '''
        Evaluate the derivative of consumption and medical care with respect to
        medical need shock at given levels of market resources, permanent income,
        and medical need shocks.

        Parameters
        ----------
        mLvl : np.array
            Market resource levels.
        pLvl : np.array
            Permanent income levels; should be same size as mLvl.
        MedShk : np.array
            Medical need shocks; should be same size as mLvl.

        Returns
        -------
        dcdShk : np.array
            Derivative of consumption with respect to medical need for each
            point in (xLvl,MedShk).
        dMeddShk : np.array
            Derivative of medical care with respect to medical need for each
            point in (xLvl,MedShk).
        '''
        xLvl = self.xFunc(mLvl,pLvl,MedShk)
        dxdShk = self.xFunc.derivativeZ(mLvl,pLvl,MedShk)
        dcdx = self.cFunc.derivativeX(xLvl,MedShk)
        dcdShk = dxdShk*dcdx + self.cFunc.derivativeY(xLvl,MedShk)
        dMeddShk = (dxdShk - dcdShk)/self.MedPrice
        return dcdShk,dMeddShk


class cThruXfunc(HARKobject):
    '''
    Class for representing consumption function derived from total expenditure
    and consumption.
    '''
    distance_criteria = ['xFunc','cFunc']

    def __init__(self,xFunc,cFunc):
        '''
        Make a new instance of MedFromXfunc.

        Parameters
        ----------
        xFunc : function
            Optimal total spending as a function of market resources, permanent
            income, and the medical need shock.
        cFunc : function
            Optimal consumption as a function of total spending and the medical
            need shock.

        Returns
        -------
        None
        '''
        self.xFunc = xFunc
        self.cFunc = cFunc

    def __call__(self,mLvl,pLvl,MedShk):
        '''
        Evaluate optimal consumption at given levels of market resources, perma-
        nent income, and medical need shocks.

        Parameters
        ----------
        mLvl : np.array
            Market resource levels.
        pLvl : np.array
            Permanent income levels; should be same size as mLvl.
        MedShk : np.array
            Medical need shocks; should be same size as mLvl.

        Returns
        -------
        cLvl : np.array
            Optimal consumption for each point in (xLvl,MedShk).
        '''
        xLvl = self.xFunc(mLvl,pLvl,MedShk)
        cLvl = self.cFunc(xLvl,MedShk)
        return cLvl

    def derivativeX(self,mLvl,pLvl,MedShk):
        '''
        Evaluate the derivative of consumption with respect to market resources
        at given levels of market resources, permanent income, and medical need
        shocks.

        Parameters
        ----------
        mLvl : np.array
            Market resource levels.
        pLvl : np.array
            Permanent income levels; should be same size as mLvl.
        MedShk : np.array
            Medical need shocks; should be same size as mLvl.

        Returns
        -------
        dcdm : np.array
            Derivative of consumption with respect to market resources for each
            point in (xLvl,MedShk).
        '''
        xLvl = self.xFunc(mLvl,pLvl,MedShk)
        dxdm = self.xFunc.derivativeX(mLvl,pLvl,MedShk)
        dcdx = self.cFunc.derivativeX(xLvl,MedShk)
        dcdm = dxdm*dcdx
        return dcdm

    def derivativeY(self,mLvl,pLvl,MedShk):
        '''
        Evaluate the derivative of consumption and medical care with respect to
        permanent income at given levels of market resources, permanent income,
        and medical need shocks.

        Parameters
        ----------
        mLvl : np.array
            Market resource levels.
        pLvl : np.array
            Permanent income levels; should be same size as mLvl.
        MedShk : np.array
            Medical need shocks; should be same size as mLvl.

        Returns
        -------
        dcdp : np.array
            Derivative of consumption with respect to permanent income for each
            point in (xLvl,MedShk).
        '''
        xLvl = self.xFunc(mLvl,pLvl,MedShk)
        dxdp = self.xFunc.derivativeY(mLvl,pLvl,MedShk)
        dcdx = self.cFunc.derivativeX(xLvl,MedShk)
        dcdp = dxdp*dcdx
        return dcdp

    def derivativeZ(self,mLvl,pLvl,MedShk):
        '''
        Evaluate the derivative of consumption and medical care with respect to
        medical need shock at given levels of market resources, permanent income,
        and medical need shocks.

        Parameters
        ----------
        mLvl : np.array
            Market resource levels.
        pLvl : np.array
            Permanent income levels; should be same size as mLvl.
        MedShk : np.array
            Medical need shocks; should be same size as mLvl.

        Returns
        -------
        dcdShk : np.array
            Derivative of consumption with respect to medical need for each
            point in (xLvl,MedShk).
        '''
        xLvl = self.xFunc(mLvl,pLvl,MedShk)
        dxdShk = self.xFunc.derivativeZ(mLvl,pLvl,MedShk)
        dcdx = self.cFunc.derivativeX(xLvl,MedShk)
        dcdShk = dxdShk*dcdx + self.cFunc.derivativeY(xLvl,MedShk)
        return dcdShk

class MedThruXfunc(HARKobject):
    '''
    Class for representing medical care function derived from total expenditure
    and consumption.
    '''
    distance_criteria = ['xFunc','cFunc','MedPrice']

    def __init__(self,xFunc,cFunc,MedPrice):
        '''
        Make a new instance of MedFromXfunc.

        Parameters
        ----------
        xFunc : function
            Optimal total spending as a function of market resources, permanent
            income, and the medical need shock.
        cFunc : function
            Optimal consumption as a function of total spending and the medical
            need shock.
        MedPrice : float
            Relative price of a unit of medical care.

        Returns
        -------
        None
        '''
        self.xFunc = xFunc
        self.cFunc = cFunc
        self.MedPrice = MedPrice

    def __call__(self,mLvl,pLvl,MedShk):
        '''
        Evaluate optimal medical care at given levels of market resources,
        permanent income, and medical need shock.

        Parameters
        ----------
        mLvl : np.array
            Market resource levels.
        pLvl : np.array
            Permanent income levels; should be same size as mLvl.
        MedShk : np.array
            Medical need shocks; should be same size as mLvl.

        Returns
        -------
        Med : np.array
            Optimal medical care for each point in (xLvl,MedShk).
        '''
        xLvl = self.xFunc(mLvl,pLvl,MedShk)
        Med  = (xLvl-self.cFunc(xLvl,MedShk))/self.MedPrice
        return Med

    def derivativeX(self,mLvl,pLvl,MedShk):
        '''
        Evaluate the derivative of consumption and medical care with respect to
        market resources at given levels of market resources, permanent income,
        and medical need shocks.

        Parameters
        ----------
        mLvl : np.array
            Market resource levels.
        pLvl : np.array
            Permanent income levels; should be same size as mLvl.
        MedShk : np.array
            Medical need shocks; should be same size as mLvl.

        Returns
        -------
        dcdm : np.array
            Derivative of consumption with respect to market resources for each
            point in (xLvl,MedShk).
        dMeddm : np.array
            Derivative of medical care with respect to market resources for each
            point in (xLvl,MedShk).
        '''
        xLvl = self.xFunc(mLvl,pLvl,MedShk)
        dxdm = self.xFunc.derivativeX(mLvl,pLvl,MedShk)
        dcdx = self.cFunc.derivativeX(xLvl,MedShk)
        dcdm = dxdm*dcdx
        dMeddm = (dxdm - dcdm)/self.MedPrice
        return dcdm,dMeddm

    def derivativeY(self,mLvl,pLvl,MedShk):
        '''
        Evaluate the derivative of medical care with respect to permanent income
        at given levels of market resources, permanent income, and medical need
        shocks.

        Parameters
        ----------
        mLvl : np.array
            Market resource levels.
        pLvl : np.array
            Permanent income levels; should be same size as mLvl.
        MedShk : np.array
            Medical need shocks; should be same size as mLvl.

        Returns
        -------
        dMeddp : np.array
            Derivative of medical care with respect to permanent income for each
            point in (xLvl,MedShk).
        '''
        xLvl = self.xFunc(mLvl,pLvl,MedShk)
        dxdp = self.xFunc.derivativeY(mLvl,pLvl,MedShk)
        dMeddp = (dxdp - dxdp*self.cFunc.derivativeX(xLvl,MedShk))/self.MedPrice
        return dMeddp

    def derivativeZ(self,mLvl,pLvl,MedShk):
        '''
        Evaluate the derivative of medical care with respect to medical need
        shock at given levels of market resources, permanent income, and medical
        need shocks.

        Parameters
        ----------
        mLvl : np.array
            Market resource levels.
        pLvl : np.array
            Permanent income levels; should be same size as mLvl.
        MedShk : np.array
            Medical need shocks; should be same size as mLvl.

        Returns
        -------
        dMeddShk : np.array
            Derivative of medical care with respect to medical need for each
            point in (xLvl,MedShk).
        '''
        xLvl = self.xFunc(mLvl,pLvl,MedShk)
        dxdShk = self.xFunc.derivativeZ(mLvl,pLvl,MedShk)
        dcdx = self.cFunc.derivativeX(xLvl,MedShk)
        dcdShk = dxdShk*dcdx + self.cFunc.derivativeY(xLvl,MedShk)
        dMeddShk = (dxdShk - dcdShk)/self.MedPrice
        return dMeddShk


###############################################################################

class MedShockConsumerType(PersistentShockConsumerType):
    '''
    A class to represent agents who consume two goods: ordinary composite consumption
    and medical care; both goods yield CRRAutility, and the coefficients on the
    goods might be different.  Agents expect to receive shocks to permanent and
    transitory income as well as multiplicative shocks to utility from medical care.
    '''
    shock_vars_ = PersistentShockConsumerType.shock_vars_ + ['MedShkNow']

    def __init__(self,cycles=1,time_flow=True,**kwds):
        '''
        Instantiate a new ConsumerType with given data, and construct objects
        to be used during solution (income distribution, assets grid, etc).
        See ConsumerParameters.init_med_shock for a dictionary of the keywords
        that should be passed to the constructor.

        Parameters
        ----------
        cycles : int
            Number of times the sequence of periods should be solved.
        time_flow : boolean
            Whether time is currently "flowing" forward for this instance.

        Returns
        -------
        None
        '''
        PersistentShockConsumerType.__init__(self,cycles=cycles,**kwds)
        self.solveOnePeriod = solveConsMedShock # Choose correct solver
        self.addToTimeInv('CRRAmed')
        self.addToTimeVary('MedPrice')

    def update(self):
        '''
        Update the income process, the assets grid, the permanent income grid,
        the medical shock distribution, and the terminal solution.

        Parameters
        ----------
        none

        Returns
        -------
        none
        '''
        self.updateIncomeProcess()
        self.updateAssetsGrid()
        self.updatepLvlNextFunc()
        self.updatepLvlGrid()
        self.updateMedShockProcess()
        self.updateSolutionTerminal()

    def updateMedShockProcess(self):
        '''
        Constructs discrete distributions of medical preference shocks for each
        period in the cycle.  Distributions are saved as attribute MedShkDstn,
        which is added to time_vary.

        Parameters
        ----------
        None

        Returns
        -------
        None
        '''
        MedShkDstn = [] # empty list for medical shock distribution each period
        for t in range(self.T_cycle):
            MedShkAvgNow  = self.MedShkAvg[t] # get shock distribution parameters
            MedShkStdNow  = self.MedShkStd[t]
            MedShkDstnNow = approxLognormal(mu=np.log(MedShkAvgNow)-0.5*MedShkStdNow**2,\
                            sigma=MedShkStdNow,N=self.MedShkCount, tail_N=self.MedShkCountTail,
                            tail_bound=[0,0.9])
            MedShkDstnNow = addDiscreteOutcomeConstantMean(MedShkDstnNow,0.0,0.0,sort=True) # add point at zero with no probability
            MedShkDstn.append(MedShkDstnNow)
        self.MedShkDstn = MedShkDstn
        self.addToTimeVary('MedShkDstn')

    def updateSolutionTerminal(self):
        '''
        Update the terminal period solution for this type.  Similar to other models,
        optimal behavior involves spending all available market resources; however,
        the agent must split his resources between consumption and medical care.

        Parameters
        ----------
        None

        Returns:
        --------
        None
        '''
        # Take last period data, whichever way time is flowing
        if self.time_flow:
            MedPrice = self.MedPrice[-1]
            MedShkVals = self.MedShkDstn[-1][1]
            MedShkPrbs = self.MedShkDstn[-1][0]
        else:
            MedPrice = self.MedPrice[0]
            MedShkVals = self.MedShkDstn[0][1]
            MedShkPrbs = self.MedShkDstn[0][0]

        # Initialize grids of medical need shocks, market resources, and optimal consumption
        MedShkGrid = MedShkVals
        xLvlMin = np.min(self.aXtraGrid)*np.min(self.pLvlGrid)
        xLvlMax = np.max(self.aXtraGrid)*np.max(self.pLvlGrid)
        xLvlGrid = makeGridExpMult(xLvlMin, xLvlMax, 3*self.aXtraGrid.size, 8)
        trivial_grid = np.array([0.0,1.0]) # Trivial grid

        # Make the policy functions for the terminal period
        xFunc_terminal = TrilinearInterp(np.array([[[0.0,0.0],[0.0,0.0]],[[1.0,1.0],[1.0,1.0]]]),\
                         trivial_grid,trivial_grid,trivial_grid)
        policyFunc_terminal = MedShockPolicyFunc(xFunc_terminal,xLvlGrid,MedShkGrid,MedPrice,
                              self.CRRA,self.CRRAmed,xLvlCubicBool=self.CubicBool)
        cFunc_terminal = cThruXfunc(xFunc_terminal,policyFunc_terminal.cFunc)
        MedFunc_terminal = MedThruXfunc(xFunc_terminal,policyFunc_terminal.cFunc,MedPrice)

        # Calculate optimal consumption on a grid of market resources and medical shocks
        mLvlGrid = xLvlGrid
        mLvlGrid_tiled = np.tile(np.reshape(mLvlGrid,(mLvlGrid.size,1)),(1,MedShkGrid.size))
        pLvlGrid_tiled = np.ones_like(mLvlGrid_tiled) # permanent income irrelevant in terminal period
        MedShkGrid_tiled = np.tile(np.reshape(MedShkVals,(1,MedShkGrid.size)),(mLvlGrid.size,1))
        cLvlGrid,MedGrid = policyFunc_terminal(mLvlGrid_tiled,pLvlGrid_tiled,MedShkGrid_tiled)

        # Integrate marginal value across shocks to get expected marginal value
        vPgrid = cLvlGrid**(-self.CRRA)
        vPgrid[np.isinf(vPgrid)] = 0.0 # correct for issue at bottom edges
        PrbGrid = np.tile(np.reshape(MedShkPrbs,(1,MedShkGrid.size)),(mLvlGrid.size,1))
        vP_expected = np.sum(vPgrid*PrbGrid,axis=1)

        # Construct the marginal (marginal) value function for the terminal period
        vPnvrs = vP_expected**(-1.0/self.CRRA)
        vPnvrs[0] = 0.0
        vPnvrsFunc = BilinearInterp(np.tile(np.reshape(vPnvrs,(vPnvrs.size,1)),
                     (1,trivial_grid.size)),mLvlGrid,trivial_grid)
        vPfunc_terminal = MargValueFunc2D(vPnvrsFunc,self.CRRA)
        vPPfunc_terminal = MargMargValueFunc2D(vPnvrsFunc,self.CRRA)

        # Integrate value across shocks to get expected value
        vGrid = utility(cLvlGrid,gam=self.CRRA) + MedShkGrid_tiled*utility(MedGrid,gam=self.CRRAmed)
        vGrid[:,0] = utility(cLvlGrid[:,0],gam=self.CRRA) # correct for issue when MedShk=0
        vGrid[np.isinf(vGrid)] = 0.0 # correct for issue at bottom edges
        v_expected = np.sum(vGrid*PrbGrid,axis=1)

        # Construct the value function for the terminal period
        vNvrs = utility_inv(v_expected,gam=self.CRRA)
        vNvrs[0] = 0.0
        vNvrsP = vP_expected*utility_invP(v_expected,gam=self.CRRA) # NEED TO FIGURE OUT MPC MAX IN THIS MODEL
        vNvrsP[0] = 0.0
        tempFunc = CubicInterp(mLvlGrid,vNvrs,vNvrsP)
        vNvrsFunc = LinearInterpOnInterp1D([tempFunc,tempFunc],trivial_grid)
        vFunc_terminal = ValueFunc2D(vNvrsFunc,self.CRRA)

        # Make the terminal period solution
        self.solution_terminal.cFunc = cFunc_terminal
        self.solution_terminal.MedFunc = MedFunc_terminal
        self.solution_terminal.policyFunc = policyFunc_terminal
        self.solution_terminal.vPfunc = vPfunc_terminal
        self.solution_terminal.vFunc = vFunc_terminal
        self.solution_terminal.vPPfunc = vPPfunc_terminal
        self.solution_terminal.hNrm = 0.0 # Don't track normalized human wealth
        self.solution_terminal.hLvl = lambda p : np.zeros_like(p) # But do track absolute human wealth by permanent income
        self.solution_terminal.mLvlMin = lambda p : np.zeros_like(p) # And minimum allowable market resources by perm inc


    def updatepLvlGrid(self):
        '''
        Update the grid of permanent income levels.  Currently only works for
        infinite horizon models (cycles=0) and lifecycle models (cycles=1).  Not
        clear what to do about cycles>1.  Identical to version in persistent
        shocks model, but pLvl=0 is manually added to the grid (because there is
        no closed form lower-bounding cFunc for pLvl=0).

        Parameters
        ----------
        None

        Returns
        -------
        None
        '''
        # Run basic version of this method
        PersistentShockConsumerType.updatepLvlGrid(self)
        for j in range(len(self.pLvlGrid)): # Then add 0 to the bottom of each pLvlGrid
            this_grid = self.pLvlGrid[j]
            self.pLvlGrid[j] = np.insert(this_grid,0,0.0001)

    def getShocks(self):
        '''
        Gets permanent and transitory income shocks for this period as well as medical need shocks
        and the price of medical care.

        Parameters
        ----------
        None

        Returns
        -------
        None
        '''
        PersistentShockConsumerType.getShocks(self) # Get permanent and transitory income shocks
        MedShkNow = np.zeros(self.AgentCount) # Initialize medical shock array
        MedPriceNow = np.zeros(self.AgentCount) # Initialize relative price array
        for t in range(self.T_cycle):
            these = t == self.t_cycle
            N = np.sum(these)
            if N > 0:
                MedShkAvg = self.MedShkAvg[t]
                MedShkStd = self.MedShkStd[t]
                MedPrice  = self.MedPrice[t]
                MedShkNow[these] = self.RNG.permutation(approxLognormal(N,mu=np.log(MedShkAvg)-0.5*MedShkStd**2,sigma=MedShkStd)[1])
                MedPriceNow[these] = MedPrice
        self.MedShkNow = MedShkNow
        self.MedPriceNow = MedPriceNow

    def getControls(self):
        '''
        Calculates consumption and medical care for each consumer of this type using the consumption
        and medical care functions.

        Parameters
        ----------
        None

        Returns
        -------
        None
        '''
        cLvlNow = np.zeros(self.AgentCount) + np.nan
        MedNow  = np.zeros(self.AgentCount) + np.nan
        for t in range(self.T_cycle):
            these = t == self.t_cycle
            cLvlNow[these], MedNow[these] = self.solution[t].policyFunc(self.mLvlNow[these],self.pLvlNow[these],self.MedShkNow[these])
        self.cLvlNow = cLvlNow
        self.MedNow  = MedNow
        return None

    def getPostStates(self):
        '''
        Calculates end-of-period assets for each consumer of this type.

        Parameters
        ----------
        None

        Returns
        -------
        None
        '''
        self.aLvlNow = self.mLvlNow - self.cLvlNow - self.MedPriceNow*self.MedNow
        return None

###############################################################################

class ConsMedShockSolver(ConsGenIncProcessSolver):
    '''
    Class for solving the one period problem for the "medical shocks" model, in
    which consumers receive shocks to permanent and transitory income as well as
    shocks to "medical need"-- multiplicative utility shocks for a second good.
    '''
    def __init__(self,solution_next,IncomeDstn,MedShkDstn,LivPrb,DiscFac,CRRA,CRRAmed,Rfree,MedPrice,
                 pLvlNextFunc,BoroCnstArt,aXtraGrid,pLvlGrid,vFuncBool,CubicBool):
        '''
        Constructor for a new solver for a one period problem with idiosyncratic
        shocks to permanent and transitory income and shocks to medical need.

        Parameters
        ----------
        solution_next : ConsumerSolution
            The solution to next period's one period problem.
        IncomeDstn : [np.array]
            A list containing three arrays of floats, representing a discrete
            approximation to the income process between the period being solved
            and the one immediately following (in solution_next). Order: event
            probabilities, permanent shocks, transitory shocks.
        MedShkDstn : [np.array]
            Discrete distribution of the multiplicative utility shifter for med-
            ical care. Order: probabilities, preference shocks.
        LivPrb : float
            Survival probability; likelihood of being alive at the beginning of
            the succeeding period.
        DiscFac : float
            Intertemporal discount factor for future utility.
        CRRA : float
            Coefficient of relative risk aversion for composite consumption.
        CRRAmed : float
            Coefficient of relative risk aversion for medical care.
        Rfree : float
            Risk free interest factor on end-of-period assets.
        MedPrice : float
            Price of unit of medical care relative to unit of consumption.
        pLvlNextFunc : float
            Expected permanent income next period as a function of current pLvl.
        BoroCnstArt: float or None
            Borrowing constraint for the minimum allowable assets to end the
            period with.
        aXtraGrid: np.array
            Array of "extra" end-of-period (normalized) asset values-- assets
            above the absolute minimum acceptable level.
        pLvlGrid: np.array
            Array of permanent income levels at which to solve the problem.
        vFuncBool: boolean
            An indicator for whether the value function should be computed and
            included in the reported solution.
        CubicBool: boolean
            An indicator for whether the solver should use cubic or linear inter-
            polation.

        Returns
        -------
        None
        '''
        ConsGenIncProcessSolver.__init__(self,solution_next,IncomeDstn,LivPrb,DiscFac,CRRA,Rfree,
                 pLvlNextFunc,BoroCnstArt,aXtraGrid,pLvlGrid,vFuncBool,CubicBool)
        self.MedShkDstn = MedShkDstn
        self.MedPrice   = MedPrice
        self.CRRAmed    = CRRAmed

    def setAndUpdateValues(self,solution_next,IncomeDstn,LivPrb,DiscFac):
        '''
        Unpacks some of the inputs (and calculates simple objects based on them),
        storing the results in self for use by other methods.  These include:
        income shocks and probabilities, medical shocks and probabilities, next
        period's marginal value function (etc), the probability of getting the
        worst income shock next period, the patience factor, human wealth, and
        the bounding MPCs.

        Parameters
        ----------
        solution_next : ConsumerSolution
            The solution to next period's one period problem.
        IncomeDstn : [np.array]
            A list containing three arrays of floats, representing a discrete
            approximation to the income process between the period being solved
            and the one immediately following (in solution_next). Order: event
            probabilities, permanent shocks, transitory shocks.
        LivPrb : float
            Survival probability; likelihood of being alive at the beginning of
            the succeeding period.
        DiscFac : float
            Intertemporal discount factor for future utility.

        Returns
        -------
        None
        '''
        # Run basic version of this method
        ConsGenIncProcessSolver.setAndUpdateValues(self,self.solution_next,self.IncomeDstn,
                                                     self.LivPrb,self.DiscFac)

        # Also unpack the medical shock distribution
        self.MedShkPrbs = self.MedShkDstn[0]
        self.MedShkVals = self.MedShkDstn[1]

    def defUtilityFuncs(self):
        '''
        Defines CRRA utility function for this period (and its derivatives,
        and their inverses), saving them as attributes of self for other methods
        to use.  Extends version from ConsIndShock models by also defining inverse
        marginal utility function over medical care.

        Parameters
        ----------
        none

        Returns
        -------
        none
        '''
        ConsGenIncProcessSolver.defUtilityFuncs(self) # Do basic version
        self.uMedPinv = lambda Med : utilityP_inv(Med,gam=self.CRRAmed)
        self.uMed     = lambda Med : utility(Med,gam=self.CRRAmed)
        self.uMedPP   = lambda Med : utilityPP(Med,gam=self.CRRAmed)

    def defBoroCnst(self,BoroCnstArt):
        '''
        Defines the constrained portion of the consumption function as cFuncNowCnst,
        an attribute of self.  Uses the artificial and natural borrowing constraints.

        Parameters
        ----------
        BoroCnstArt : float or None
            Borrowing constraint for the minimum allowable (normalized) assets
            to end the period with.  If it is less than the natural borrowing
            constraint at a particular permanent income level, then it is irrelevant;
            BoroCnstArt=None indicates no artificial borrowing constraint.

        Returns
        -------
        None
        '''
        # Find minimum allowable end-of-period assets at each permanent income level
        PermIncMinNext = self.PermShkMinNext*self.pLvlNextFunc(self.pLvlGrid)
        IncLvlMinNext  = PermIncMinNext*self.TranShkMinNext
        aLvlMin = (self.solution_next.mLvlMin(PermIncMinNext) - IncLvlMinNext)/self.Rfree

        # Make a function for the natural borrowing constraint by permanent income
        BoroCnstNat = LinearInterp(np.insert(self.pLvlGrid,0,0.0),np.insert(aLvlMin,0,0.0))
        self.BoroCnstNat = BoroCnstNat

        # Define the minimum allowable level of market resources by permanent income
        if self.BoroCnstArt is not None:
            BoroCnstArt = LinearInterp([0.0,1.0],[0.0,self.BoroCnstArt])
            self.mLvlMinNow = UpperEnvelope(BoroCnstNat,BoroCnstArt)
        else:
            self.mLvlMinNow = BoroCnstNat

        # Make the constrained total spending function: spend all market resources
        trivial_grid = np.array([0.0,1.0]) # Trivial grid
        spendAllFunc = TrilinearInterp(np.array([[[0.0,0.0],[0.0,0.0]],[[1.0,1.0],[1.0,1.0]]]),\
                       trivial_grid,trivial_grid,trivial_grid)
        self.xFuncNowCnst = VariableLowerBoundFunc3D(spendAllFunc,self.mLvlMinNow)

        self.mNrmMinNow = 0.0 # Needs to exist so as not to break when solution is created
        self.MPCmaxEff  = 0.0 # Actually might vary by p, but no use formulating as a function

    def getPointsForInterpolation(self,EndOfPrdvP,aLvlNow):
        '''
        Finds endogenous interpolation points (x,m) for the expenditure function.

        Parameters
        ----------
        EndOfPrdvP : np.array
            Array of end-of-period marginal values.
        aLvlNow : np.array
            Array of end-of-period asset values that yield the marginal values
            in EndOfPrdvP.

        Returns
        -------
        x_for_interpolation : np.array
            Total expenditure points for interpolation.
        m_for_interpolation : np.array
            Corresponding market resource points for interpolation.
        p_for_interpolation : np.array
            Corresponding permanent income points for interpolation.
        '''
        # Get size of each state dimension
        mCount      = aLvlNow.shape[1]
        pCount      = aLvlNow.shape[0]
        MedCount = self.MedShkVals.size

        # Calculate endogenous gridpoints and controls
        cLvlNow = np.tile(np.reshape(self.uPinv(EndOfPrdvP),(1,pCount,mCount)),(MedCount,1,1))
        MedBaseNow = np.tile(np.reshape(self.uMedPinv(self.MedPrice*EndOfPrdvP),(1,pCount,mCount)),
                             (MedCount,1,1))
        MedShkVals_tiled = np.tile(np.reshape(self.MedShkVals**(1.0/self.CRRAmed),(MedCount,1,1)),
                                   (1,pCount,mCount))
        MedLvlNow = MedShkVals_tiled*MedBaseNow
        aLvlNow_tiled = np.tile(np.reshape(aLvlNow,(1,pCount,mCount)),(MedCount,1,1))
        xLvlNow = cLvlNow + self.MedPrice*MedLvlNow
        mLvlNow = xLvlNow + aLvlNow_tiled

        # Limiting consumption is zero as m approaches the natural borrowing constraint
        x_for_interpolation = np.concatenate((np.zeros((MedCount,pCount,1)),xLvlNow),axis=-1)
        temp = np.tile(self.BoroCnstNat(np.reshape(self.pLvlGrid,(1,self.pLvlGrid.size,1))),
                       (MedCount,1,1))
        m_for_interpolation = np.concatenate((temp,mLvlNow),axis=-1)

        # Make a 3D array of permanent income for interpolation
        p_for_interpolation = np.tile(np.reshape(self.pLvlGrid,(1,pCount,1)),(MedCount,1,mCount+1))

        # Store for use by cubic interpolator
        self.cLvlNow = cLvlNow
        self.MedLvlNow = MedLvlNow
        self.MedShkVals_tiled = np.tile(np.reshape(self.MedShkVals,(MedCount,1,1)),(1,pCount,mCount))

        return x_for_interpolation, m_for_interpolation, p_for_interpolation

    def usePointsForInterpolation(self,xLvl,mLvl,pLvl,MedShk,interpolator):
        '''
        Constructs a basic solution for this period, including the consumption
        function and marginal value function.

        Parameters
        ----------
        xLvl : np.array
            Total expenditure points for interpolation.
        mLvl : np.array
            Corresponding market resource points for interpolation.
        pLvl : np.array
            Corresponding permanent income level points for interpolation.
        MedShk : np.array
            Corresponding medical need shocks for interpolation.
        interpolator : function
            A function that constructs and returns a consumption function.

        Returns
        -------
        solution_now : ConsumerSolution
            The solution to this period's consumption-saving problem, with a
            consumption function, marginal value function, and minimum m.
        '''
        # Construct the unconstrained total expenditure function
        xFuncNowUnc = interpolator(mLvl,pLvl,MedShk,xLvl)
        xFuncNowCnst = self.xFuncNowCnst
        xFuncNow = LowerEnvelope3D(xFuncNowUnc,xFuncNowCnst)

        # Transform the expenditure function into policy functions for consumption and medical care
        aug_factor = 2
        xLvlGrid = makeGridExpMult(np.min(xLvl),np.max(xLvl),aug_factor*self.aXtraGrid.size,8)
        policyFuncNow = MedShockPolicyFunc(xFuncNow,xLvlGrid,self.MedShkVals,self.MedPrice,
                        self.CRRA,self.CRRAmed,xLvlCubicBool=self.CubicBool)
        cFuncNow = cThruXfunc(xFuncNow,policyFuncNow.cFunc)
        MedFuncNow = MedThruXfunc(xFuncNow,policyFuncNow.cFunc,self.MedPrice)

        # Make the marginal value function (and the value function if vFuncBool=True)
        vFuncNow, vPfuncNow = self.makevAndvPfuncs(policyFuncNow)

        # Pack up the solution and return it
        solution_now = ConsumerSolution(cFunc=cFuncNow, vFunc=vFuncNow, vPfunc=vPfuncNow,\
                                        mNrmMin=self.mNrmMinNow)
        solution_now.MedFunc = MedFuncNow
        solution_now.policyFunc = policyFuncNow
        return solution_now

    def makevAndvPfuncs(self,policyFunc):
        '''
        Constructs the marginal value function for this period.

        Parameters
        ----------
        policyFunc : function
            Consumption and medical care function for this period, defined over
            market resources, permanent income level, and the medical need shock.

        Returns
        -------
        vFunc : function
            Value function for this period, defined over market resources and
            permanent income.
        vPfunc : function
            Marginal value (of market resources) function for this period, defined
            over market resources and permanent income.
        '''
        # Get state dimension sizes
        mCount   = self.aXtraGrid.size
        pCount   = self.pLvlGrid.size
        MedCount = self.MedShkVals.size

        # Make temporary grids to evaluate the consumption function
        temp_grid  = np.tile(np.reshape(self.aXtraGrid,(mCount,1,1)),(1,pCount,MedCount))
        aMinGrid   = np.tile(np.reshape(self.mLvlMinNow(self.pLvlGrid),(1,pCount,1)),
                             (mCount,1,MedCount))
        pGrid      = np.tile(np.reshape(self.pLvlGrid,(1,pCount,1)),(mCount,1,MedCount))
        mGrid      = temp_grid*pGrid + aMinGrid
        if self.pLvlGrid[0] == 0:
            mGrid[:,0,:] = np.tile(np.reshape(self.aXtraGrid,(mCount,1)),(1,MedCount))
        MedShkGrid = np.tile(np.reshape(self.MedShkVals,(1,1,MedCount)),(mCount,pCount,1))
        probsGrid  = np.tile(np.reshape(self.MedShkPrbs,(1,1,MedCount)),(mCount,pCount,1))

        # Get optimal consumption (and medical care) for each state
        cGrid,MedGrid = policyFunc(mGrid,pGrid,MedShkGrid)

        # Calculate expected value by "integrating" across medical shocks
        if self.vFuncBool:
            MedGrid = np.maximum(MedGrid,1e-100) # interpolation error sometimes makes Med < 0 (barely)
            aGrid = np.maximum(mGrid - cGrid - self.MedPrice*MedGrid, aMinGrid) # interpolation error sometimes makes tiny violations
            vGrid = self.u(cGrid) + MedShkGrid*self.uMed(MedGrid) + self.EndOfPrdvFunc(aGrid,pGrid)
            vNow  = np.sum(vGrid*probsGrid,axis=2)

        # Calculate expected marginal value by "integrating" across medical shocks
        vPgrid = self.uP(cGrid)
        vPnow  = np.sum(vPgrid*probsGrid,axis=2)

        # Add vPnvrs=0 at m=mLvlMin to close it off at the bottom (and vNvrs=0)
        mGrid_small = np.concatenate((np.reshape(self.mLvlMinNow(self.pLvlGrid),(1,pCount)),mGrid[:,:,0]))
        vPnvrsNow   = np.concatenate((np.zeros((1,pCount)),self.uPinv(vPnow)))
        if self.vFuncBool:
            vNvrsNow  = np.concatenate((np.zeros((1,pCount)),self.uinv(vNow)),axis=0)
            vNvrsPnow = vPnow*self.uinvP(vNow)
            vNvrsPnow = np.concatenate((np.zeros((1,pCount)),vNvrsPnow),axis=0)

        # Construct the pseudo-inverse value and marginal value functions over mLvl,pLvl
        vPnvrsFunc_by_pLvl = []
        vNvrsFunc_by_pLvl = []
        for j in range(pCount): # Make a pseudo inverse marginal value function for each pLvl
            pLvl = self.pLvlGrid[j]
            m_temp = mGrid_small[:,j] - self.mLvlMinNow(pLvl)
            vPnvrs_temp = vPnvrsNow[:,j]
            vPnvrsFunc_by_pLvl.append(LinearInterp(m_temp,vPnvrs_temp))
            if self.vFuncBool:
                vNvrs_temp  = vNvrsNow[:,j]
                vNvrsP_temp = vNvrsPnow[:,j]
                vNvrsFunc_by_pLvl.append(CubicInterp(m_temp,vNvrs_temp,vNvrsP_temp))
        vPnvrsFuncBase = LinearInterpOnInterp1D(vPnvrsFunc_by_pLvl,self.pLvlGrid)
        vPnvrsFunc = VariableLowerBoundFunc2D(vPnvrsFuncBase,self.mLvlMinNow) # adjust for the lower bound of mLvl
        if self.vFuncBool:
            vNvrsFuncBase  = LinearInterpOnInterp1D(vNvrsFunc_by_pLvl,self.pLvlGrid)
            vNvrsFunc = VariableLowerBoundFunc2D(vNvrsFuncBase,self.mLvlMinNow) # adjust for the lower bound of mLvl

        # "Re-curve" the (marginal) value function
        vPfunc = MargValueFunc2D(vPnvrsFunc,self.CRRA)
        if self.vFuncBool:
            vFunc = ValueFunc2D(vNvrsFunc,self.CRRA)
        else:
            vFunc = NullFunc()

        return vFunc, vPfunc

    def makeLinearxFunc(self,mLvl,pLvl,MedShk,xLvl):
        '''
        Constructs the (unconstrained) expenditure function for this period using
        bilinear interpolation (over permanent income and the medical shock) among
        an array of linear interpolations over market resources.

        Parameters
        ----------
        mLvl : np.array
            Corresponding market resource points for interpolation.
        pLvl : np.array
            Corresponding permanent income level points for interpolation.
        MedShk : np.array
            Corresponding medical need shocks for interpolation.
        xLvl : np.array
            Expenditure points for interpolation, corresponding to those in mLvl,
            pLvl, and MedShk.

        Returns
        -------
        xFuncUnc : BilinearInterpOnInterp1D
            Unconstrained total expenditure function for this period.
        '''
        # Get state dimensions
        pCount = mLvl.shape[1]
        MedCount = mLvl.shape[0]

        # Loop over each permanent income level and medical shock and make a linear xFunc
        xFunc_by_pLvl_and_MedShk = [] # Initialize the empty list of lists of 1D xFuncs
        for i in range(pCount):
            temp_list = []
            pLvl_i = pLvl[0,i,0]
            mLvlMin_i = self.BoroCnstNat(pLvl_i)
            for j in range(MedCount):
                m_temp = mLvl[j,i,:] - mLvlMin_i
                x_temp = xLvl[j,i,:]
                temp_list.append(LinearInterp(m_temp,x_temp))
            xFunc_by_pLvl_and_MedShk.append(deepcopy(temp_list))

        # Combine the nested list of linear xFuncs into a single function
        pLvl_temp = pLvl[0,:,0]
        MedShk_temp = MedShk[:,0,0]
        xFuncUncBase = BilinearInterpOnInterp1D(xFunc_by_pLvl_and_MedShk,pLvl_temp,MedShk_temp)
        xFuncUnc = VariableLowerBoundFunc3D(xFuncUncBase,self.BoroCnstNat)
        return xFuncUnc

    def makeCubicxFunc(self,mLvl,pLvl,MedShk,xLvl):
        '''
        Constructs the (unconstrained) expenditure function for this period using
        bilinear interpolation (over permanent income and the medical shock) among
        an array of cubic interpolations over market resources.

        Parameters
        ----------
        mLvl : np.array
            Corresponding market resource points for interpolation.
        pLvl : np.array
            Corresponding permanent income level points for interpolation.
        MedShk : np.array
            Corresponding medical need shocks for interpolation.
        xLvl : np.array
            Expenditure points for interpolation, corresponding to those in mLvl,
            pLvl, and MedShk.

        Returns
        -------
        xFuncUnc : BilinearInterpOnInterp1D
            Unconstrained total expenditure function for this period.
        '''
        # Get state dimensions
        pCount = mLvl.shape[1]
        MedCount = mLvl.shape[0]

        # Calculate the MPC and MPM at each gridpoint
        EndOfPrdvPP = self.DiscFacEff*self.Rfree*self.Rfree*np.sum(self.vPPfuncNext(self.mLvlNext,\
                      self.pLvlNext)*self.ShkPrbs_temp,axis=0)
        EndOfPrdvPP = np.tile(np.reshape(EndOfPrdvPP,(1,pCount,EndOfPrdvPP.shape[1])),(MedCount,1,1))
        dcda        = EndOfPrdvPP/self.uPP(np.array(self.cLvlNow))
        dMedda      = EndOfPrdvPP/(self.MedShkVals_tiled*self.uMedPP(self.MedLvlNow))
        dMedda[0,:,:] = 0.0 # dMedda goes crazy when MedShk=0
        MPC         = dcda/(1.0 + dcda + self.MedPrice*dMedda)
        MPM         = dMedda/(1.0 + dcda + self.MedPrice*dMedda)

        # Convert to marginal propensity to spend
        MPX = MPC + self.MedPrice*MPM
        MPX = np.concatenate((np.reshape(MPX[:,:,0],(MedCount,pCount,1)),MPX),axis=2) # NEED TO CALCULATE MPM AT NATURAL BORROWING CONSTRAINT
        MPX[0,:,0] = self.MPCmaxNow

        # Loop over each permanent income level and medical shock and make a cubic xFunc
        xFunc_by_pLvl_and_MedShk = [] # Initialize the empty list of lists of 1D xFuncs
        for i in range(pCount):
            temp_list = []
            pLvl_i = pLvl[0,i,0]
            mLvlMin_i = self.BoroCnstNat(pLvl_i)
            for j in range(MedCount):
                m_temp = mLvl[j,i,:] - mLvlMin_i
                x_temp = xLvl[j,i,:]
                MPX_temp = MPX[j,i,:]
                temp_list.append(CubicInterp(m_temp,x_temp,MPX_temp))
            xFunc_by_pLvl_and_MedShk.append(deepcopy(temp_list))

        # Combine the nested list of cubic xFuncs into a single function
        pLvl_temp = pLvl[0,:,0]
        MedShk_temp = MedShk[:,0,0]
        xFuncUncBase = BilinearInterpOnInterp1D(xFunc_by_pLvl_and_MedShk,pLvl_temp,MedShk_temp)
        xFuncUnc = VariableLowerBoundFunc3D(xFuncUncBase,self.BoroCnstNat)
        return xFuncUnc

    def makeBasicSolution(self,EndOfPrdvP,aLvl,interpolator):
        '''
        Given end of period assets and end of period marginal value, construct
        the basic solution for this period.

        Parameters
        ----------
        EndOfPrdvP : np.array
            Array of end-of-period marginal values.
        aLvl : np.array
            Array of end-of-period asset values that yield the marginal values
            in EndOfPrdvP.
        interpolator : function
            A function that constructs and returns a consumption function.

        Returns
        -------
        solution_now : ConsumerSolution
            The solution to this period's consumption-saving problem, with a
            consumption function, marginal value function, and minimum m.
        '''
        xLvl,mLvl,pLvl = self.getPointsForInterpolation(EndOfPrdvP,aLvl)
        MedShk_temp    = np.tile(np.reshape(self.MedShkVals,(self.MedShkVals.size,1,1)),\
                         (1,mLvl.shape[1],mLvl.shape[2]))
        solution_now   = self.usePointsForInterpolation(xLvl,mLvl,pLvl,MedShk_temp,interpolator)
        return solution_now

    def addvPPfunc(self,solution):
        '''
        Adds the marginal marginal value function to an existing solution, so
        that the next solver can evaluate vPP and thus use cubic interpolation.

        Parameters
        ----------
        solution : ConsumerSolution
            The solution to this single period problem, which must include the
            consumption function.

        Returns
        -------
        solution : ConsumerSolution
            The same solution passed as input, but with the marginal marginal
            value function for this period added as the attribute vPPfunc.
        '''
        vPPfuncNow        = MargMargValueFunc2D(solution.vPfunc.cFunc,self.CRRA)
        solution.vPPfunc  = vPPfuncNow
        return solution

    def solve(self):
        '''
        Solves a one period consumption saving problem with risky income and
        shocks to medical need.

        Parameters
        ----------
        None

        Returns
        -------
        solution : ConsumerSolution
            The solution to the one period problem, including a consumption
            function, medical spending function ( both defined over market re-
            sources, permanent income, and medical shock), a marginal value func-
            tion (defined over market resources and permanent income), and human
            wealth as a function of permanent income.
        '''
        aLvl,trash  = self.prepareToCalcEndOfPrdvP()
        EndOfPrdvP = self.calcEndOfPrdvP()
        if self.vFuncBool:
            self.makeEndOfPrdvFunc(EndOfPrdvP)
        if self.CubicBool:
            interpolator = self.makeCubicxFunc
        else:
            interpolator = self.makeLinearxFunc
        solution   = self.makeBasicSolution(EndOfPrdvP,aLvl,interpolator)
        solution   = self.addMPCandHumanWealth(solution)
        if self.CubicBool:
            solution = self.addvPPfunc(solution)
        return solution


def solveConsMedShock(solution_next,IncomeDstn,MedShkDstn,LivPrb,DiscFac,CRRA,CRRAmed,Rfree,MedPrice,
                 pLvlNextFunc,BoroCnstArt,aXtraGrid,pLvlGrid,vFuncBool,CubicBool):
    '''
    Solve the one period problem for a consumer with shocks to permanent and
    transitory income as well as medical need shocks (as multiplicative shifters
    for utility from a second medical care good).

    Parameters
    ----------
    solution_next : ConsumerSolution
        The solution to next period's one period problem.
    IncomeDstn : [np.array]
        A list containing three arrays of floats, representing a discrete
        approximation to the income process between the period being solved
        and the one immediately following (in solution_next). Order: event
        probabilities, permanent shocks, transitory shocks.
    MedShkDstn : [np.array]
        Discrete distribution of the multiplicative utility shifter for med-
        ical care. Order: probabilities, preference shocks.
    LivPrb : float
        Survival probability; likelihood of being alive at the beginning of
        the succeeding period.
    DiscFac : float
        Intertemporal discount factor for future utility.
    CRRA : float
        Coefficient of relative risk aversion for composite consumption.
    CRRAmed : float
        Coefficient of relative risk aversion for medical care.
    Rfree : float
        Risk free interest factor on end-of-period assets.
    MedPrice : float
        Price of unit of medical care relative to unit of consumption.
    pLvlNextFunc : float
        Expected permanent income next period as a function of current pLvl.
    PrstIncCorr : float
        Correlation of permanent income from period to period.
    BoroCnstArt: float or None
        Borrowing constraint for the minimum allowable assets to end the
        period with.
    aXtraGrid: np.array
        Array of "extra" end-of-period (normalized) asset values-- assets
        above the absolute minimum acceptable level.
    pLvlGrid: np.array
        Array of permanent income levels at which to solve the problem.
    vFuncBool: boolean
        An indicator for whether the value function should be computed and
        included in the reported solution.
    CubicBool: boolean
        An indicator for whether the solver should use cubic or linear inter-
        polation.

    Returns
    -------
    solution : ConsumerSolution
        Solution to this period's problem, including a consumption function,
        medical spending function, and marginal value function.  The former two
        are defined over (mLvl,pLvl,MedShk), while the latter is defined only
        on (mLvl,pLvl), with MedShk integrated out.
    '''
    solver = ConsMedShockSolver(solution_next,IncomeDstn,MedShkDstn,LivPrb,DiscFac,CRRA,CRRAmed,Rfree,
                MedPrice,pLvlNextFunc,BoroCnstArt,aXtraGrid,pLvlGrid,vFuncBool,CubicBool)
    solver.prepareToSolve()       # Do some preparatory work
    solution_now = solver.solve() # Solve the period
    return solution_now


###############################################################################

if __name__ == '__main__':
    import ConsumerParameters as Params
    from HARK.utilities import CRRAutility_inv
    from time import clock
    import matplotlib.pyplot as plt
    mystr = lambda number : "{:.4f}".format(number)

    do_simulation = True

    # Make and solve an example medical shocks consumer type
    MedicalExample = MedShockConsumerType(**Params.init_medical_shocks)
    t_start = clock()
    MedicalExample.solve()
    t_end = clock()
    print('Solving a medical shocks consumer took ' + mystr(t_end-t_start) + ' seconds.')

    # Plot the consumption function
    M = np.linspace(0,30,300)
    pLvl = 1.0
    P = pLvl*np.ones_like(M)
    for j in range(MedicalExample.MedShkDstn[0][0].size):
        MedShk = MedicalExample.MedShkDstn[0][1][j]*np.ones_like(M)
        M_temp = M + MedicalExample.solution[0].mLvlMin(pLvl)
        C = MedicalExample.solution[0].cFunc(M_temp,P,MedShk)
        plt.plot(M_temp,C)
    print('Consumption function by medical need shock (constant permanent income)')
    plt.show()

    # Plot the medical care function
    for j in range(MedicalExample.MedShkDstn[0][0].size):
        MedShk = MedicalExample.MedShkDstn[0][1][j]*np.ones_like(M)
        Med = MedicalExample.solution[0].MedFunc(M_temp,P,MedShk)
        plt.plot(M_temp,Med)
    print('Medical care function by medical need shock (constant permanent income)')
    plt.ylim([0,20])
    plt.show()

    # Plot the savings function
    for j in range(MedicalExample.MedShkDstn[0][0].size):
        MedShk = MedicalExample.MedShkDstn[0][1][j]*np.ones_like(M)
        Sav = M_temp - MedicalExample.solution[0].cFunc(M_temp,P,MedShk) - MedicalExample.MedPrice[0]*\
              MedicalExample.solution[0].MedFunc(M_temp,P,MedShk)
        plt.plot(M_temp,Sav)
    print('End of period savings by medical need shock (constant permanent income)')
    plt.show()

    # Plot the marginal value function
    M = np.linspace(0.0,30,300)
    for p in range(MedicalExample.pLvlGrid[0].size):
        pLvl = MedicalExample.pLvlGrid[0][p]
        M_temp = pLvl*M + MedicalExample.solution[0].mLvlMin(pLvl)
        P = pLvl*np.ones_like(M)
        vP = MedicalExample.solution[0].vPfunc(M_temp,P)**(-1.0/MedicalExample.CRRA)
        plt.plot(M_temp,vP)
    print('Marginal value function (pseudo inverse)')
    plt.show()

    if MedicalExample.vFuncBool:
        # Plot the value function
        M = np.linspace(0.0,1,300)
        for p in range(MedicalExample.pLvlGrid[0].size):
            pLvl = MedicalExample.pLvlGrid[0][p]
            M_temp = pLvl*M + MedicalExample.solution[0].mLvlMin(pLvl)
            P = pLvl*np.ones_like(M)
            v = CRRAutility_inv(MedicalExample.solution[0].vFunc(M_temp,P),gam=MedicalExample.CRRA)
            plt.plot(M_temp,v)
        print('Value function (pseudo inverse)')
        plt.show()

    if do_simulation:
        t_start = clock()
        MedicalExample.T_sim = 100
        MedicalExample.track_vars = ['mLvlNow','cLvlNow','MedNow']
        MedicalExample.makeShockHistory()
        MedicalExample.initializeSim()
        MedicalExample.simulate()
        t_end = clock()
        print('Simulating ' + str(MedicalExample.AgentCount) + ' agents for ' + str(MedicalExample.T_sim) + ' periods took ' + mystr(t_end-t_start) + ' seconds.')



'''
This module makes some figures for cstwMPC.  It requires that quite a few specifications
of the model have been estimated, with the results stored in ./Results.
'''

import matplotlib.pyplot as plt
import csv
import numpy as np

f = open('./Results/LCbetaPointNetWorthLorenzFig.txt','r')
my_reader = csv.reader(f,delimiter='\t')
raw_data = list(my_reader)
lorenz_percentiles = []
scf_lorenz = []
beta_point_lorenz = []
for j in range(len(raw_data)):
    lorenz_percentiles.append(float(raw_data[j][0]))
    scf_lorenz.append(float(raw_data[j][1]))
    beta_point_lorenz.append(float(raw_data[j][2]))
f.close()
lorenz_percentiles = np.array(lorenz_percentiles)
scf_lorenz = np.array(scf_lorenz)
beta_point_lorenz = np.array(beta_point_lorenz)

f = open('./Results/LCbetaDistNetWorthLorenzFig.txt','r')
my_reader = csv.reader(f,delimiter='\t')
raw_data = list(my_reader)
beta_dist_lorenz = []
for j in range(len(raw_data)):
    beta_dist_lorenz.append(float(raw_data[j][2]))
f.close()
beta_dist_lorenz = np.array(beta_dist_lorenz)

f = open('./Results/LCbetaPointNetWorthMPCfig.txt','r')
my_reader = csv.reader(f,delimiter='\t')
raw_data = list(my_reader)
mpc_percentiles = []
mpc_beta_point = []
for j in range(len(raw_data)):
    mpc_percentiles.append(float(raw_data[j][0]))
    mpc_beta_point.append(float(raw_data[j][1]))
f.close()
mpc_percentiles = np.asarray(mpc_percentiles)
mpc_beta_point = np.asarray(mpc_beta_point)

f = open('./Results/LCbetaDistNetWorthMPCfig.txt','r')
my_reader = csv.reader(f,delimiter='\t')
raw_data = list(my_reader)
mpc_beta_dist = []
for j in range(len(raw_data)):
    mpc_beta_dist.append(float(raw_data[j][1]))
f.close()
mpc_beta_dist = np.asarray(mpc_beta_dist)

f = open('./Results/LCbetaDistLiquidMPCfig.txt','r')
my_reader = csv.reader(f,delimiter='\t')
raw_data = list(my_reader)
mpc_beta_dist_liquid = []
for j in range(len(raw_data)):
    mpc_beta_dist_liquid.append(float(raw_data[j][1]))
f.close()
mpc_beta_dist_liquid = np.asarray(mpc_beta_dist_liquid)

f = open('./Results/LCbetaDistNetWorthKappaByAge.txt','r')
my_reader = csv.reader(f,delimiter='\t')
raw_data = list(my_reader)
kappa_mean_age = []
kappa_lo_beta_age = []
kappa_hi_beta_age = []
for j in range(len(raw_data)):
    kappa_mean_age.append(float(raw_data[j][0]))
    kappa_lo_beta_age.append(float(raw_data[j][1]))
    kappa_hi_beta_age.append(float(raw_data[j][2]))
kappa_mean_age = np.array(kappa_mean_age)
kappa_lo_beta_age = np.array(kappa_lo_beta_age)
kappa_hi_beta_age = np.array(kappa_hi_beta_age)
age_list = np.array(range(len(kappa_mean_age)),dtype=float)*0.25 + 24.0
f.close()

f = open('./Results/LC_KYbyBeta.txt','r')
my_reader = csv.reader(f,delimiter='\t')
raw_data = list(my_reader)
KY_by_beta_lifecycle = []
beta_list = []
for j in range(len(raw_data)):
    beta_list.append(float(raw_data[j][0]))
    KY_by_beta_lifecycle.append(float(raw_data[j][1]))
beta_list = np.array(beta_list)
KY_by_beta_lifecycle = np.array(KY_by_beta_lifecycle)
f.close()

f = open('./Results/IH_KYbyBeta.txt','r')
my_reader = csv.reader(f,delimiter='\t')
raw_data = list(my_reader)
KY_by_beta_infinite = []
for j in range(len(raw_data)):
    KY_by_beta_infinite.append(float(raw_data[j][1]))
KY_by_beta_infinite = np.array(KY_by_beta_infinite)
f.close()


plt.plot(100*lorenz_percentiles,beta_point_lorenz,'-.k',linewidth=1.5)
plt.plot(100*lorenz_percentiles,beta_dist_lorenz,'--k',linewidth=1.5)
plt.plot(100*lorenz_percentiles,scf_lorenz,'-k',linewidth=1.5)
plt.xlabel('Wealth percentile',fontsize=14)
plt.ylabel('Cumulative wealth ownership',fontsize=14)
plt.title('Lorenz Curve Matching, Lifecycle Model',fontsize=16)
plt.legend((r'$\beta$-point',r'$\beta$-dist','SCF data'),loc=2,fontsize=12)
plt.ylim(-0.01,1)
plt.savefig('./Figures/LorenzLifecycle.pdf')
plt.show()

plt.plot(mpc_beta_point,mpc_percentiles,'-.k',linewidth=1.5)
plt.plot(mpc_beta_dist,mpc_percentiles,'--k',linewidth=1.5)
plt.plot(mpc_beta_dist_liquid,mpc_percentiles,'-.k',linewidth=1.5)
plt.xlabel('Marginal propensity to consume',fontsize=14)
plt.ylabel('Cumulative probability',fontsize=14)
plt.title('CDF of the MPC, Lifecycle Model',fontsize=16)
plt.legend((r'$\beta$-point NW',r'$\beta$-dist NW',r'$\beta$-dist LA'),loc=0,fontsize=12)
plt.savefig('./Figures/MPCdistLifecycle.pdf')
plt.show()

plt.plot(age_list,kappa_mean_age,'-k',linewidth=1.5)
plt.plot(age_list,kappa_lo_beta_age,'--k',linewidth=1.5)
plt.plot(age_list,kappa_hi_beta_age,'-.k',linewidth=1.5)
plt.legend(('Population average','Most impatient','Most patient'),loc=2,fontsize=12)
plt.xlabel('Age',fontsize=14)
plt.ylabel('Average MPC',fontsize=14)
plt.title('Marginal Propensity to Consume by Age',fontsize=16)
plt.xlim(24,100)
plt.ylim(0,1)
plt.savefig('./Figures/MPCbyAge.pdf')
plt.show()

plt.plot(beta_list,KY_by_beta_infinite,'-k',linewidth=1.5)
plt.plot(beta_list,KY_by_beta_lifecycle,'--k',linewidth=1.5)
plt.plot([0.95,1.01],[10.26,10.26],'--k',linewidth=0.75)
plt.text(0.96,12,'U.S. K/Y ratio')
plt.legend(('Perpetual youth','Lifecycle'),loc=2,fontsize=12)
plt.xlabel(r'Discount factor $\beta$',fontsize=14)
plt.ylabel('Capital to output ratio',fontsize=14)
plt.title('K/Y Ratio by Discount Factor',fontsize=16)
plt.ylim(0,100)
plt.savefig('./Figures/KYratioByBeta.pdf')
plt.show()


f = open('./Results/IHbetaPointNetWorthLorenzFig.txt','r')
my_reader = csv.reader(f,delimiter='\t')
raw_data = list(my_reader)
lorenz_percentiles = []
scf_lorenz = []
beta_point_lorenz = []
for j in range(len(raw_data)):
    lorenz_percentiles.append(float(raw_data[j][0]))
    scf_lorenz.append(float(raw_data[j][1]))
    beta_point_lorenz.append(float(raw_data[j][2]))
f.close()
lorenz_percentiles = np.array(lorenz_percentiles)
scf_lorenz = np.array(scf_lorenz)
beta_point_lorenz = np.array(beta_point_lorenz)

f = open('./Results/IHbetaDistNetWorthLorenzFig.txt','r')
my_reader = csv.reader(f,delimiter='\t')
raw_data = list(my_reader)
beta_dist_lorenz = []
for j in range(len(raw_data)):
    beta_dist_lorenz.append(float(raw_data[j][2]))
f.close()
beta_dist_lorenz = np.array(beta_dist_lorenz)


f = open('./Results/IHbetaPointLiquidLorenzFig.txt','r')
my_reader = csv.reader(f,delimiter='\t')
raw_data = list(my_reader)
beta_point_lorenz_liquid = []
for j in range(len(raw_data)):
    beta_point_lorenz_liquid.append(float(raw_data[j][2]))
f.close()
beta_point_lorenz_liquid = np.array(beta_point_lorenz_liquid)

f = open('./Results/IHbetaDistLiquidLorenzFig.txt','r')
my_reader = csv.reader(f,delimiter='\t')
raw_data = list(my_reader)
beta_dist_lorenz_liquid = []
for j in range(len(raw_data)):
    beta_dist_lorenz_liquid.append(float(raw_data[j][2]))
f.close()
beta_dist_lorenz_liquid = np.array(beta_dist_lorenz_liquid)

f = open('./Results/IHbetaPointNetWorthMPCfig.txt','r')
my_reader = csv.reader(f,delimiter='\t')
raw_data = list(my_reader)
mpc_percentiles = []
mpc_beta_point = []
for j in range(len(raw_data)):
    mpc_percentiles.append(float(raw_data[j][0]))
    mpc_beta_point.append(float(raw_data[j][1]))
f.close()
mpc_percentiles = np.asarray(mpc_percentiles)
mpc_beta_point = np.asarray(mpc_beta_point)

f = open('./Results/IHbetaDistNetWorthMPCfig.txt','r')
my_reader = csv.reader(f,delimiter='\t')
raw_data = list(my_reader)
mpc_beta_dist = []
for j in range(len(raw_data)):
    mpc_beta_dist.append(float(raw_data[j][1]))
f.close()
mpc_beta_dist = np.asarray(mpc_beta_dist)

f = open('./Results/IHbetaDistLiquidMPCfig.txt','r')
my_reader = csv.reader(f,delimiter='\t')
raw_data = list(my_reader)
mpc_beta_dist_liquid = []
for j in range(len(raw_data)):
    mpc_beta_dist_liquid.append(float(raw_data[j][1]))
f.close()
mpc_beta_dist_liquid = np.asarray(mpc_beta_dist_liquid)


plt.plot(100*lorenz_percentiles,beta_point_lorenz,'-.k',linewidth=1.5)
plt.plot(100*lorenz_percentiles,scf_lorenz,'-k',linewidth=1.5)
plt.xlabel('Wealth percentile',fontsize=14)
plt.ylabel('Cumulative wealth ownership',fontsize=14)
plt.title('Lorenz Curve Matching, Perpetual Youth Model',fontsize=16)
plt.legend((r'$\beta$-point','SCF data'),loc=2,fontsize=12)
plt.ylim(-0.01,1)
plt.savefig('./Figures/LorenzInfiniteBP.pdf')
plt.show()

plt.plot(100*lorenz_percentiles,beta_point_lorenz,'-.k',linewidth=1.5)
plt.plot(100*lorenz_percentiles,beta_dist_lorenz,'--k',linewidth=1.5)
plt.plot(100*lorenz_percentiles,scf_lorenz,'-k',linewidth=1.5)
plt.xlabel('Wealth percentile',fontsize=14)
plt.ylabel('Cumulative wealth ownership',fontsize=14)
plt.title('Lorenz Curve Matching, Perpetual Youth Model',fontsize=16)
plt.legend((r'$\beta$-point',r'$\beta$-dist','SCF data'),loc=2,fontsize=12)
plt.ylim(-0.01,1)
plt.savefig('./Figures/LorenzInfinite.pdf')
plt.show()

plt.plot(100*lorenz_percentiles,beta_point_lorenz_liquid,'-.k',linewidth=1.5)
plt.plot(100*lorenz_percentiles,beta_dist_lorenz_liquid,'--k',linewidth=1.5)
plt.plot(np.array([20,40,60,80]),np.array([0.0, 0.004, 0.025,0.117]),'.r',markersize=10)
plt.xlabel('Wealth percentile',fontsize=14)
plt.ylabel('Cumulative wealth ownership',fontsize=14)
plt.title('Lorenz Curve Matching, Perpetual Youth (Liquid Assets)',fontsize=16)
plt.legend((r'$\beta$-point',r'$\beta$-dist','SCF targets'),loc=2,fontsize=12)
plt.ylim(-0.01,1)
plt.savefig('./Figures/LorenzLiquid.pdf')
plt.show()

plt.plot(mpc_beta_point,mpc_percentiles,'-.k',linewidth=1.5)
plt.plot(mpc_beta_dist,mpc_percentiles,'--k',linewidth=1.5)
plt.plot(mpc_beta_dist_liquid,mpc_percentiles,'-.k',linewidth=1.5)
plt.xlabel('Marginal propensity to consume',fontsize=14)
plt.ylabel('Cumulative probability',fontsize=14)
plt.title('CDF of the MPC, Perpetual Youth Model',fontsize=16)
plt.legend((r'$\beta$-point NW',r'$\beta$-dist NW',r'$\beta$-dist LA'),loc=0,fontsize=12)
plt.savefig('./Figures/MPCdistInfinite.pdf')
plt.show()




f = open('./Results/SensitivityRho.txt','r')
my_reader = csv.reader(f,delimiter='\t')
raw_data = list(my_reader)
rho_sensitivity = np.array(raw_data)
f.close()

f = open('./Results/SensitivityXiSigma.txt','r')
my_reader = csv.reader(f,delimiter='\t')
raw_data = list(my_reader)
xi_sigma_sensitivity = np.array(raw_data)
f.close()

f = open('./Results/SensitivityPsiSigma.txt','r')
my_reader = csv.reader(f,delimiter='\t')
raw_data = list(my_reader)
psi_sigma_sensitivity = np.array(raw_data)
f.close()

f = open('./Results/SensitivityMu.txt','r')
my_reader = csv.reader(f,delimiter='\t')
raw_data = list(my_reader)
mu_sensitivity = np.array(raw_data)
f.close()

f = open('./Results/SensitivityUrate.txt','r')
my_reader = csv.reader(f,delimiter='\t')
raw_data = list(my_reader)
urate_sensitivity = np.array(raw_data)
f.close()

f = open('./Results/SensitivityMortality.txt','r')
my_reader = csv.reader(f,delimiter='\t')
raw_data = list(my_reader)
mortality_sensitivity = np.array(raw_data)
f.close()

f = open('./Results/SensitivityG.txt','r')
my_reader = csv.reader(f,delimiter='\t')
raw_data = list(my_reader)
g_sensitivity = np.array(raw_data)
f.close()


#plt.subplots(2,3,sharey=True)
kappa = 0.242
plt.subplot(2,3,1,xticks=[0.5,2.3,4],ylim=(0.21,0.36))
plt.plot(rho_sensitivity[:,0],rho_sensitivity[:,1],'-k',linewidth=1)
plt.plot(1,kappa,'.k',markersize=8)
plt.yticks([0.22,0.26,0.30,0.34])
plt.text(0.7,0.215,r'Risk aversion $\rho$',fontsize=11)
plt.ylabel('Aggregate MPC',fontsize=11)

plt.subplot(2,3,2,xticks=[0.0,0.4,0.8],xlim=(0,0.8),ylim=(0.21,0.36))
plt.plot(xi_sigma_sensitivity[:,0],xi_sigma_sensitivity[:,1],'-k',linewidth=1)
plt.plot(0.2,kappa,'.k',markersize=8)
plt.yticks([0.22,0.26,0.30,0.34],[])
plt.text(0.045,0.215,r'Transitory std $\sigma_\theta$',fontsize=11)
plt.title('Sensitivity Analysis: Perpetual Youth',fontsize=14)

plt.subplot(2,3,3,xticks=[0.04,0.06,0.08],xlim=(0.04,0.08),ylim=(0.21,0.36))
plt.plot(psi_sigma_sensitivity[:,0],psi_sigma_sensitivity[:,1],'-k',linewidth=1)
plt.plot(0.0603,kappa,'.k',markersize=8)
plt.yticks([0.22,0.26,0.30,0.34],[])
plt.text(0.041,0.34,r'Permanent std $\sigma_\psi$',fontsize=11)

plt.subplot(2,3,4,xticks=[0.0,0.4,0.8],ylim=(0.21,0.36))
plt.plot(mu_sensitivity[:,0],mu_sensitivity[:,1],'-k',linewidth=1)
plt.plot(0.15,kappa,'.k',markersize=8)
plt.yticks([0.22,0.26,0.30,0.34])
plt.text(0.05,0.34,'Unemployment',fontsize=11)
plt.text(0.22,0.32,r'benefit $\mu$',fontsize=11)
plt.ylabel('Aggregate MPC',fontsize=11)

plt.subplot(2,3,5,xticks=[0.02,0.07,0.12],ylim=(0.21,0.36))
plt.plot(urate_sensitivity[:,0],urate_sensitivity[:,1],'-k',linewidth=1)
plt.plot(0.07,kappa,'.k',markersize=8)
plt.yticks([0.22,0.26,0.30,0.34],[])
plt.text(0.03,0.34,'Unemployment',fontsize=11)
plt.text(0.055,0.32,r'rate $\mho$',fontsize=11)

'''
plt.subplot(2,3,6,xticks=[0.004,0.008,0.012],xlim=(0.003,0.0125),ylim=(0.21,0.36))
plt.plot(mortality_sensitivity[:,0],mortality_sensitivity[:,1],'-k',linewidth=1)
plt.plot(0.00625,kappa,'.k',markersize=8)
plt.yticks([0.22,0.26,0.30,0.34],[])
plt.text(0.0037,0.34,r'Mortality rate $\mathsf{D}$',fontsize=11)
'''
plt.subplot(2,3,6,xlim=(0.0,0.04),ylim=(0.21,0.36))
plt.plot(g_sensitivity[:,0],g_sensitivity[:,1],'-k',linewidth=1)
plt.plot(0.01,kappa,'.k',markersize=8)
plt.xticks([0.0,0.02,0.04],['0','0.02','0.04'])
plt.yticks([0.22,0.26,0.30,0.34],[])
plt.text(0.008,0.34,r'Aggregate',fontsize=11)
plt.text(0.005,0.32,r'growth rate $g$',fontsize=11)

plt.savefig('./Figures/KappaSensitivity.pdf')
plt.show()



#plt.subplots(2,3,sharey=True)
beta = 0.9877
plt.subplot(2,3,1,xticks=[0.5,2.3,4],ylim=(0.95,1.0))
plt.plot(rho_sensitivity[:,0],rho_sensitivity[:,2],'-k',linewidth=1)
plt.plot(1,beta,'.k',markersize=8)
plt.yticks([0.95,0.96,0.97,0.98,0.99,1.0])
plt.text(0.7,0.955,r'Risk aversion $\rho$',fontsize=11)
plt.ylabel(r'Estimated $\grave{\beta}$',fontsize=11)

plt.subplot(2,3,2,xticks=[0.0,0.4,0.8],xlim=(0,0.8),ylim=(0.95,1.0))
plt.plot(xi_sigma_sensitivity[:,0],xi_sigma_sensitivity[:,2],'-k',linewidth=1)
plt.plot(0.2,beta,'.k',markersize=8)
plt.yticks([0.95,0.96,0.97,0.98,0.99,1.0],[])
plt.text(0.045,0.955,r'Transitory std $\sigma_\theta$',fontsize=11)
plt.title('Sensitivity Analysis: Perpetual Youth',fontsize=14)

plt.subplot(2,3,3,xticks=[0.04,0.06,0.08],xlim=(0.04,0.08),ylim=(0.97,1.0))
plt.plot(psi_sigma_sensitivity[:,0],psi_sigma_sensitivity[:,2],'-k',linewidth=1)
plt.plot(0.0603,beta,'.k',markersize=8)
plt.yticks([0.95,0.96,0.97,0.98,0.99,1.0],[])
plt.text(0.041,0.955,r'Permanent std $\sigma_\psi$',fontsize=11)

plt.subplot(2,3,4,xticks=[0.0,0.4,0.8],ylim=(0.95,1.0))
plt.plot(mu_sensitivity[:,0],mu_sensitivity[:,2],'-k',linewidth=1)
plt.plot(0.15,beta,'.k',markersize=8)
plt.yticks([0.95,0.96,0.97,0.98,0.99,1.0])
plt.text(0.05,0.9625,'Unemployment',fontsize=11)
plt.text(0.22,0.955,r'benefit $\mu$',fontsize=11)
plt.ylabel(r'Estimated $\grave{\beta}$',fontsize=11)

plt.subplot(2,3,5,xticks=[0.02,0.07,0.12],ylim=(0.95,1.0))
plt.plot(urate_sensitivity[:,0],urate_sensitivity[:,2],'-k',linewidth=1)
plt.plot(0.07,beta,'.k',markersize=8)
plt.yticks([0.95,0.96,0.97,0.98,0.99,1.0],[])
plt.text(0.03,0.9625,'Unemployment',fontsize=11)
plt.text(0.055,0.955,r'rate $\mho$',fontsize=11)

'''
plt.subplot(2,3,6,xticks=[0.004,0.008,0.012],xlim=(0.003,0.0125),ylim=(0.95,1.0))
plt.plot(mortality_sensitivity[:,0],mortality_sensitivity[:,2],'-k',linewidth=1)
plt.plot(0.00625,beta,'.k',markersize=8)
plt.yticks([0.95,0.96,0.97,0.98,0.99,1.0],[])
plt.text(0.0037,0.955,r'Mortality rate $\mathsf{D}$',fontsize=11)
'''
plt.subplot(2,3,6,xlim=(0.0,0.04),ylim=(0.95,1.0))
plt.plot(g_sensitivity[:,0],g_sensitivity[:,2],'-k',linewidth=1)
plt.plot(0.01,beta,'.k',markersize=8)
plt.xticks([0.0,0.02,0.04],['0','0.02','0.04'])
plt.yticks([0.95,0.96,0.97,0.98,0.99,1.0],[])
plt.text(0.008,0.9625,r'Aggregate',fontsize=11)
plt.text(0.005,0.955,r'growth rate $g$',fontsize=11)

plt.savefig('./Figures/BetaSensitivity.pdf')
plt.show()



#plt.subplots(2,3,sharey=True)
nabla = 0.00736
plt.subplot(2,3,1,xticks=[0.5,2.3,4],ylim=(0.000,0.055))
plt.plot(rho_sensitivity[:,0],rho_sensitivity[:,3],'-k',linewidth=1)
plt.plot(1,nabla,'.k',markersize=8)
plt.yticks([0,0.01,0.02,0.03,0.04,0.05])
plt.text(0.63,0.0475,r'Risk aversion $\rho$',fontsize=11)
plt.ylabel(r'Estimated $\nabla$',fontsize=11)

plt.subplot(2,3,2,xticks=[0.0,0.4,0.8],xlim=(0,0.8),ylim=(0.000,0.055))
plt.plot(xi_sigma_sensitivity[:,0],xi_sigma_sensitivity[:,3],'-k',linewidth=1)
plt.plot(0.2,nabla,'.k',markersize=8)
plt.yticks([0,0.01,0.02,0.03,0.04,0.05],[])
plt.text(0.045,0.0475,r'Transitory std $\sigma_\theta$',fontsize=11)
plt.title('Sensitivity Analysis: Perpetual Youth',fontsize=14)

plt.subplot(2,3,3,xticks=[0.04,0.06,0.08],xlim=(0.04,0.08),ylim=(0.00,0.055))
plt.plot(psi_sigma_sensitivity[:,0],psi_sigma_sensitivity[:,3],'-k',linewidth=1)
plt.plot(0.0603,nabla,'.k',markersize=8)
plt.yticks([0,0.01,0.02,0.03,0.04,0.05],[])
plt.text(0.041,0.0475,r'Permanent std $\sigma_\psi$',fontsize=11)

plt.subplot(2,3,4,xticks=[0.0,0.4,0.8],ylim=(0.000,0.055))
plt.plot(mu_sensitivity[:,0],mu_sensitivity[:,3],'-k',linewidth=1)
plt.plot(0.15,nabla,'.k',markersize=8)
plt.yticks([0,0.01,0.02,0.03,0.04,0.05])
plt.text(0.05,0.0475,'Unemployment',fontsize=11)
plt.text(0.22,0.040,r'benefit $\mu$',fontsize=11)
plt.ylabel(r'Estimated $\nabla$',fontsize=11)

plt.subplot(2,3,5,xticks=[0.02,0.07,0.12],ylim=(0.000,0.055))
plt.plot(urate_sensitivity[:,0],urate_sensitivity[:,3],'-k',linewidth=1)
plt.plot(0.07,nabla,'.k',markersize=8)
plt.yticks([0,0.01,0.02,0.03,0.04,0.05],[])
plt.text(0.03,0.0475,'Unemployment',fontsize=11)
plt.text(0.055,0.04,r'rate $\mho$',fontsize=11)

'''
plt.subplot(2,3,6,xticks=[0.004,0.008,0.012],xlim=(0.003,0.0125),ylim=(0.000,0.055))
plt.plot(mortality_sensitivity[:,0],mortality_sensitivity[:,3],'-k',linewidth=1)
plt.plot(0.00625,nabla,'.k',markersize=8)
plt.yticks([0,0.01,0.02,0.03,0.04,0.05],[])
plt.text(0.0037,0.0475,r'Mortality rate $\mathsf{D}$',fontsize=11)
'''

plt.subplot(2,3,6,xlim=(0.0,0.04),ylim=(0.000,0.055))
plt.plot(g_sensitivity[:,0],g_sensitivity[:,3],'-k',linewidth=1)
plt.plot(0.01,nabla,'.k',markersize=8)
plt.xticks([0.0,0.02,0.04],['0','0.02','0.04'])
plt.yticks([0,0.01,0.02,0.03,0.04,0.05],[])
plt.text(0.008,0.0475,r'Aggregate',fontsize=11)
plt.text(0.005,0.04,r'growth rate $g$',fontsize=11)

plt.savefig('./Figures/NablaSensitivity.pdf')
plt.show()



#plt.subplots(2,3,sharey=True)
fit = 4.593
plt.subplot(2,3,1,xticks=[0.5,2.3,4],ylim=(0,10))
plt.plot(rho_sensitivity[:,0],rho_sensitivity[:,4],'-k',linewidth=1)
plt.plot(1,fit,'.k',markersize=8)
plt.yticks([1,3,5,7,9])
plt.text(0.7,0.5,r'Risk aversion $\rho$',fontsize=11)
plt.ylabel('Lorenz distance',fontsize=11)

plt.subplot(2,3,2,xticks=[0.0,0.4,0.8],xlim=(0,0.8),ylim=(0,10))
plt.plot(xi_sigma_sensitivity[:,0],xi_sigma_sensitivity[:,4],'-k',linewidth=1)
plt.plot(0.2,fit,'.k',markersize=8)
plt.yticks([1,3,5,7,9],[])
plt.text(0.05,0.5,r'Transitory std $\sigma_\theta$',fontsize=11)
plt.title('Sensitivity Analysis: Perpetual Youth',fontsize=14)

plt.subplot(2,3,3,xticks=[0.04,0.06,0.08],xlim=(0.04,0.08),ylim=(0,10))
plt.plot(psi_sigma_sensitivity[:,0],psi_sigma_sensitivity[:,4],'-k',linewidth=1)
plt.plot(0.0603,fit,'.k',markersize=8)
plt.yticks([1,3,5,7,9],[])
plt.text(0.041,0.5,r'Permanent std $\sigma_\psi$',fontsize=11)

plt.subplot(2,3,4,xticks=[0.0,0.4,0.8],ylim=(0,10))
plt.plot(mu_sensitivity[:,0],mu_sensitivity[:,4],'-k',linewidth=1)
plt.plot(0.15,fit,'.k',markersize=8)
plt.yticks([1,3,5,7,9])
plt.text(0.05,8.5,'Unemployment',fontsize=11)
plt.text(0.22,7.25,r'benefit $\mu$',fontsize=11)
plt.ylabel('Lorenz distance',fontsize=11)

plt.subplot(2,3,5,xticks=[0.02,0.07,0.12],ylim=(0,10))
plt.plot(urate_sensitivity[:,0],urate_sensitivity[:,4],'-k',linewidth=1)
plt.plot(0.07,fit,'.k',markersize=8)
plt.yticks([1,3,5,7,9],[])
plt.text(0.03,8.5,'Unemployment',fontsize=11)
plt.text(0.055,7.25,r'rate $\mho$',fontsize=11)

'''
plt.subplot(2,3,6,xticks=[0.004,0.008,0.012],xlim=(0.003,0.0125),ylim=(0,10))
plt.plot(mortality_sensitivity[:,0],mortality_sensitivity[:,4],'-k',linewidth=1)
plt.plot(0.00625,fit,'.k',markersize=8)
plt.yticks([1,3,5,7,9],[])
plt.text(0.0037,0.5,r'Mortality rate $\mathsf{D}$',fontsize=11)
'''

plt.subplot(2,3,6,xlim=(0.0,0.04),ylim=(0,10))
plt.plot(g_sensitivity[:,0],g_sensitivity[:,4],'-k',linewidth=1)
plt.plot(0.01,fit,'.k',markersize=8)
plt.xticks([0.0,0.02,0.04],['0','0.02','0.04'])
plt.yticks([1,3,5,7,9],[])
plt.text(0.008,8.5,r'Aggregate',fontsize=11)
plt.text(0.005,7.25,r'growth rate $g$',fontsize=11)

plt.savefig('./Figures/FitSensitivity.pdf')
plt.show()



'''
A second stab / complete do-over of cstwMPC.  Steals some bits from old version.
'''

import numpy as np
from copy import copy, deepcopy
from time import clock
from HARK.utilities import approxMeanOneLognormal, combineIndepDstns, approxUniform, \
                           getPercentiles, getLorenzShares, calcSubpopAvg, approxLognormal
from HARK.simulation import drawDiscrete
from HARK import Market
import SetupParamsCSTW as Params
import HARK.ConsumptionSaving.ConsIndShockModel as Model
from HARK.ConsumptionSaving.ConsAggShockModel import CobbDouglasEconomy, AggShockConsumerType
from scipy.optimize import golden, brentq
import matplotlib.pyplot as plt
#import csv

mystr = lambda number : "{:.3f}".format(number)

if Params.do_agg_shocks:
    EstimationAgentClass = AggShockConsumerType
    EstimationMarketClass = CobbDouglasEconomy
else:
    EstimationAgentClass = Model.IndShockConsumerType
    EstimationMarketClass = Market

class cstwMPCagent(EstimationAgentClass):
    '''
    A slight extension of the idiosyncratic consumer type for the cstwMPC model.
    '''
    def reset(self):
        self.initializeSim()
        self.t_age = drawDiscrete(self.AgentCount,P=self.AgeDstn,X=np.arange(self.AgeDstn.size),exact_match=False,seed=self.RNG.randint(0,2**31-1)).astype(int)
        self.t_cycle = copy(self.t_age)
        if hasattr(self,'kGrid'):
            self.aLvlNow = self.kInit*np.ones(self.AgentCount) # Start simulation near SS
            self.aNrmNow = self.aLvlNow/self.pLvlNow

    def marketAction(self):
        if hasattr(self,'kGrid'):
            self.pLvl = self.pLvlNow/np.mean(self.pLvlNow)
        self.simulate(1)

    def updateIncomeProcess(self):
        '''
        An alternative method for constructing the income process in the infinite horizon model.

        Parameters
        ----------
        none

        Returns
        -------
        none
        '''
        if self.cycles == 0:
            tax_rate = (self.IncUnemp*self.UnempPrb)/((1.0-self.UnempPrb)*self.IndL)
            TranShkDstn     = deepcopy(approxMeanOneLognormal(self.TranShkCount,sigma=self.TranShkStd[0],tail_N=0))
            TranShkDstn[0]  = np.insert(TranShkDstn[0]*(1.0-self.UnempPrb),0,self.UnempPrb)
            TranShkDstn[1]  = np.insert(TranShkDstn[1]*(1.0-tax_rate)*self.IndL,0,self.IncUnemp)
            PermShkDstn     = approxMeanOneLognormal(self.PermShkCount,sigma=self.PermShkStd[0],tail_N=0)
            self.IncomeDstn = [combineIndepDstns(PermShkDstn,TranShkDstn)]
            self.TranShkDstn = TranShkDstn
            self.PermShkDstn = PermShkDstn
            self.addToTimeVary('IncomeDstn')
        else: # Do the usual method if this is the lifecycle model
            EstimationAgentClass.updateIncomeProcess(self)

class cstwMPCmarket(EstimationMarketClass):
    '''
    A class for representing the economy in the cstwMPC model.
    '''
    reap_vars = ['aLvlNow','pLvlNow','MPCnow','TranShkNow','EmpNow','t_age']
    sow_vars  = [] # Nothing needs to be sent back to agents in the idiosyncratic shocks version
    const_vars = ['LorenzBool','ManyStatsBool']
    track_vars = ['MaggNow','AaggNow','KtoYnow','Lorenz','LorenzLong','MPCall','MPCretired','MPCemployed','MPCunemployed','MPCbyIncome','MPCbyWealthRatio','HandToMouthPct']
    dyn_vars = [] # No dynamics in the idiosyncratic shocks version

    def __init__(self,**kwds):
        '''
        Make a new instance of cstwMPCmarket.
        '''
        self.assignParameters(**kwds)
        if self.AggShockBool:
            self.sow_vars=['MaggNow','AaggNow','RfreeNow','wRteNow','PermShkAggNow','TranShkAggNow','KtoLnow']
            self.dyn_vars=['AFunc']
            self.max_loops = 20

    def solve(self):
        '''
        Solves the cstwMPCmarket.
        '''
        if self.AggShockBool:
            for agent in self.agents:
                agent.getEconomyData(self)
            Market.solve(self)
        else:
            self.solveAgents()
            self.makeHistory()

    def millRule(self,aLvlNow,pLvlNow,MPCnow,TranShkNow,EmpNow,t_age,LorenzBool,ManyStatsBool):
        '''
        The millRule for this class simply calls the method calcStats.
        '''
        self.calcStats(aLvlNow,pLvlNow,MPCnow,TranShkNow,EmpNow,t_age,LorenzBool,ManyStatsBool)
        if self.AggShockBool:
            return self.calcRandW(aLvlNow,pLvlNow)
        else: # These variables are tracked but not created in no-agg-shocks specifications
            self.MaggNow = 0.0
            self.AaggNow = 0.0

    def calcStats(self,aLvlNow,pLvlNow,MPCnow,TranShkNow,EmpNow,t_age,LorenzBool,ManyStatsBool):
        '''
        Calculate various statistics about the current population in the economy.

        Parameters
        ----------
        aLvlNow : [np.array]
            Arrays with end-of-period assets, listed by each ConsumerType in self.agents.
        pLvlNow : [np.array]
            Arrays with permanent income levels, listed by each ConsumerType in self.agents.
        MPCnow : [np.array]
            Arrays with marginal propensity to consume, listed by each ConsumerType in self.agents.
        TranShkNow : [np.array]
            Arrays with transitory income shocks, listed by each ConsumerType in self.agents.
        EmpNow : [np.array]
            Arrays with employment states: True if employed, False otherwise.
        t_age : [np.array]
            Arrays with periods elapsed since model entry, listed by each ConsumerType in self.agents.
        LorenzBool: bool
            Indicator for whether the Lorenz target points should be calculated.  Usually False,
            only True when DiscFac has been identified for a particular nabla.
        ManyStatsBool: bool
            Indicator for whether a lot of statistics for tables should be calculated. Usually False,
            only True when parameters have been estimated and we want values for tables.

        Returns
        -------
        None
        '''
        # Combine inputs into single arrays
        aLvl = np.hstack(aLvlNow)
        pLvl = np.hstack(pLvlNow)
        age  = np.hstack(t_age)
        TranShk = np.hstack(TranShkNow)
        Emp = np.hstack(EmpNow)

        # Calculate the capital to income ratio in the economy
        CohortWeight = self.PopGroFac**(-age)
        CapAgg = np.sum(aLvl*CohortWeight)
        IncAgg = np.sum(pLvl*TranShk*CohortWeight)
        KtoYnow = CapAgg/IncAgg
        self.KtoYnow = KtoYnow

        # Store Lorenz data if requested
        self.LorenzLong = np.nan
        if LorenzBool:
            order = np.argsort(aLvl)
            aLvl = aLvl[order]
            CohortWeight = CohortWeight[order]
            wealth_shares = getLorenzShares(aLvl,weights=CohortWeight,percentiles=self.LorenzPercentiles,presorted=True)
            self.Lorenz = wealth_shares
            if ManyStatsBool:
                self.LorenzLong = getLorenzShares(aLvl,weights=CohortWeight,percentiles=np.arange(0.01,1.0,0.01),presorted=True)
        else:
            self.Lorenz = np.nan # Store nothing if we don't want Lorenz data

        # Calculate a whole bunch of statistics if requested
        if ManyStatsBool:
            # Reshape other inputs
            MPC  = np.hstack(MPCnow)

            # Sort other data items if aLvl and CohortWeight were sorted
            if LorenzBool:
                pLvl = pLvl[order]
                MPC  = MPC[order]
                TranShk = TranShk[order]
                age = age[order]
                Emp = Emp[order]
            aNrm = aLvl/pLvl # Normalized assets (wealth ratio)
            IncLvl = TranShk*pLvl # Labor income this period

            # Calculate overall population MPC and by subpopulations
            MPCannual = 1.0 - (1.0 - MPC)**4
            self.MPCall = np.sum(MPCannual*CohortWeight)/np.sum(CohortWeight)
            employed =  Emp
            unemployed = np.logical_not(employed)
            if self.T_retire > 0: # Adjust for the lifecycle model, where agents might be retired instead
                unemployed = np.logical_and(unemployed,age < self.T_retire)
                employed   = np.logical_and(employed,age < self.T_retire)
                retired    = age >= self.T_retire
            else:
                retired    = np.zeros_like(unemployed,dtype=bool)
            self.MPCunemployed = np.sum(MPCannual[unemployed]*CohortWeight[unemployed])/np.sum(CohortWeight[unemployed])
            self.MPCemployed   = np.sum(MPCannual[employed]*CohortWeight[employed])/np.sum(CohortWeight[employed])
            self.MPCretired    = np.sum(MPCannual[retired]*CohortWeight[retired])/np.sum(CohortWeight[retired])
            self.MPCbyWealthRatio = calcSubpopAvg(MPCannual,aNrm,self.cutoffs,CohortWeight)
            self.MPCbyIncome      = calcSubpopAvg(MPCannual,IncLvl,self.cutoffs,CohortWeight)

            # Calculate the wealth quintile distribution of "hand to mouth" consumers
            quintile_cuts = getPercentiles(aLvl,weights=CohortWeight,percentiles=[0.2, 0.4, 0.6, 0.8])
            wealth_quintiles = np.ones(aLvl.size,dtype=int)
            wealth_quintiles[aLvl > quintile_cuts[0]] = 2
            wealth_quintiles[aLvl > quintile_cuts[1]] = 3
            wealth_quintiles[aLvl > quintile_cuts[2]] = 4
            wealth_quintiles[aLvl > quintile_cuts[3]] = 5
            MPC_cutoff = getPercentiles(MPCannual,weights=CohortWeight,percentiles=[2.0/3.0]) # Looking at consumers with MPCs in the top 1/3
            these = MPCannual > MPC_cutoff
            in_top_third_MPC = wealth_quintiles[these]
            temp_weights = CohortWeight[these]
            hand_to_mouth_total = np.sum(temp_weights)
            hand_to_mouth_pct = []
            for q in range(1,6):
                hand_to_mouth_pct.append(np.sum(temp_weights[in_top_third_MPC == q])/hand_to_mouth_total)
            self.HandToMouthPct = np.array(hand_to_mouth_pct)

        else: # If we don't want these stats, just put empty values in history
            self.MPCall = np.nan
            self.MPCunemployed = np.nan
            self.MPCemployed = np.nan
            self.MPCretired = np.nan
            self.MPCbyWealthRatio = np.nan
            self.MPCbyIncome = np.nan
            self.HandToMouthPct = np.nan

    def distributeParams(self,param_name,param_count,center,spread,dist_type):
        '''
        Distributes heterogeneous values of one parameter to the AgentTypes in self.agents.

        Parameters
        ----------
        param_name : string
            Name of the parameter to be assigned.
        param_count : int
            Number of different values the parameter will take on.
        center : float
            A measure of centrality for the distribution of the parameter.
        spread : float
            A measure of spread or diffusion for the distribution of the parameter.
        dist_type : string
            The type of distribution to be used.  Can be "lognormal" or "uniform" (can expand).

        Returns
        -------
        None
        '''
        # Get a list of discrete values for the parameter
        if dist_type == 'uniform':
            # If uniform, center is middle of distribution, spread is distance to either edge
            param_dist = approxUniform(N=param_count,bot=center-spread,top=center+spread)
        elif dist_type == 'lognormal':
            # If lognormal, center is the mean and spread is the standard deviation (in log)
            tail_N = 3
            param_dist = approxLognormal(N=param_count-tail_N,mu=np.log(center)-0.5*spread**2,sigma=spread,tail_N=tail_N,tail_bound=[0.0,0.9], tail_order=np.e)

        # Distribute the parameters to the various types, assigning consecutive types the same
        # value if there are more types than values
        replication_factor = len(self.agents)/param_count
        j = 0
        b = 0
        while j < len(self.agents):
            for n in range(replication_factor):
                self.agents[j](AgentCount = int(self.Population*param_dist[0][b]*self.TypeWeight[n]))
                exec('self.agents[j](' + param_name + '= param_dist[1][b])')
                j += 1
            b += 1

    def calcKYratioDifference(self):
        '''
        Returns the difference between the simulated capital to income ratio and the target ratio.
        Can only be run after solving all AgentTypes and running makeHistory.

        Parameters
        ----------
        None

        Returns
        -------
        diff : float
            Difference between simulated and target capital to income ratio.
        '''
        # Ignore the first X periods to allow economy to stabilize from initial conditions
        KYratioSim = np.mean(np.array(self.KtoYnow_hist)[self.ignore_periods:])
        diff = KYratioSim - self.KYratioTarget
        return diff

    def calcLorenzDistance(self):
        '''
        Returns the sum of squared differences between simulated and target Lorenz points.

        Parameters
        ----------
        None

        Returns
        -------
        dist : float
            Sum of squared distances between simulated and target Lorenz points (sqrt)
        '''
        LorenzSim = np.mean(np.array(self.Lorenz_hist)[self.ignore_periods:,:],axis=0)
        dist = np.sqrt(np.sum((100*(LorenzSim - self.LorenzTarget))**2))
        self.LorenzDistance = dist
        return dist

    def showManyStats(self,spec_name=None):
        '''
        Calculates the "many statistics" by averaging histories across simulated periods.  Displays
        the results as text and saves them to files if spec_name is not None.

        Parameters
        ----------
        spec_name : string
            A name or label for the current specification.

        Returns
        -------
        None
        '''
        # Calculate MPC overall and by subpopulations
        MPCall = np.mean(self.MPCall_hist[self.ignore_periods:])
        MPCemployed = np.mean(self.MPCemployed_hist[self.ignore_periods:])
        MPCunemployed = np.mean(self.MPCunemployed_hist[self.ignore_periods:])
        MPCretired = np.mean(self.MPCretired_hist[self.ignore_periods:])
        MPCbyIncome = np.mean(np.array(self.MPCbyIncome_hist)[self.ignore_periods:,:],axis=0)
        MPCbyWealthRatio = np.mean(np.array(self.MPCbyWealthRatio_hist)[self.ignore_periods:,:],axis=0)
        HandToMouthPct = np.mean(np.array(self.HandToMouthPct_hist)[self.ignore_periods:,:],axis=0)

        LorenzSim = np.hstack((np.array(0.0),np.mean(np.array(self.LorenzLong_hist)[self.ignore_periods:,:],axis=0),np.array(1.0)))
        LorenzAxis = np.arange(101,dtype=float)
        plt.plot(LorenzAxis,self.LorenzData,'-k',linewidth=1.5)
        plt.plot(LorenzAxis,LorenzSim,'--k',linewidth=1.5)
        plt.xlabel('Income percentile',fontsize=12)
        plt.ylabel('Cumulative wealth share',fontsize=12)
        plt.ylim([-0.02,1.0])
        plt.show()

        # Make a string of results to display
        results_string = 'Estimate is center=' + str(self.center_estimate) + ', spread=' + str(self.spread_estimate) + '\n'
        results_string += 'Lorenz distance is ' + str(self.LorenzDistance) + '\n'
        results_string += 'Average MPC for all consumers is ' + mystr(MPCall) + '\n'
        results_string += 'Average MPC in the top percentile of W/Y is ' + mystr(MPCbyWealthRatio[0]) + '\n'
        results_string += 'Average MPC in the top decile of W/Y is ' + mystr(MPCbyWealthRatio[1]) + '\n'
        results_string += 'Average MPC in the top quintile of W/Y is ' + mystr(MPCbyWealthRatio[2]) + '\n'
        results_string += 'Average MPC in the second quintile of W/Y is ' + mystr(MPCbyWealthRatio[3]) + '\n'
        results_string += 'Average MPC in the middle quintile of W/Y is ' + mystr(MPCbyWealthRatio[4]) + '\n'
        results_string += 'Average MPC in the fourth quintile of W/Y is ' + mystr(MPCbyWealthRatio[5]) + '\n'
        results_string += 'Average MPC in the bottom quintile of W/Y is ' + mystr(MPCbyWealthRatio[6]) + '\n'
        results_string += 'Average MPC in the top percentile of y is ' + mystr(MPCbyIncome[0]) + '\n'
        results_string += 'Average MPC in the top decile of y is ' + mystr(MPCbyIncome[1]) + '\n'
        results_string += 'Average MPC in the top quintile of y is ' + mystr(MPCbyIncome[2]) + '\n'
        results_string += 'Average MPC in the second quintile of y is ' + mystr(MPCbyIncome[3]) + '\n'
        results_string += 'Average MPC in the middle quintile of y is ' + mystr(MPCbyIncome[4]) + '\n'
        results_string += 'Average MPC in the fourth quintile of y is ' + mystr(MPCbyIncome[5]) + '\n'
        results_string += 'Average MPC in the bottom quintile of y is ' + mystr(MPCbyIncome[6]) + '\n'
        results_string += 'Average MPC for the employed is ' + mystr(MPCemployed) + '\n'
        results_string += 'Average MPC for the unemployed is ' + mystr(MPCunemployed) + '\n'
        results_string += 'Average MPC for the retired is ' + mystr(MPCretired) + '\n'
        results_string += 'Of the population with the 1/3 highest MPCs...' + '\n'
        results_string += mystr(HandToMouthPct[0]*100) + '% are in the bottom wealth quintile,' + '\n'
        results_string += mystr(HandToMouthPct[1]*100) + '% are in the second wealth quintile,' + '\n'
        results_string += mystr(HandToMouthPct[2]*100) + '% are in the third wealth quintile,' + '\n'
        results_string += mystr(HandToMouthPct[3]*100) + '% are in the fourth wealth quintile,' + '\n'
        results_string += 'and ' + mystr(HandToMouthPct[4]*100) + '% are in the top wealth quintile.' + '\n'
        print(results_string)

        # Save results to disk
        if spec_name is not None:
            with open('./Results/' + spec_name + 'Results.txt','w') as f:
                f.write(results_string)
                f.close()

def getKYratioDifference(Economy,param_name,param_count,center,spread,dist_type):
    '''
    Finds the difference between simulated and target capital to income ratio in an economy when
    a given parameter has heterogeneity according to some distribution.

    Parameters
    ----------
    Economy : cstwMPCmarket
        An object representing the entire economy, containing the various AgentTypes as an attribute.
    param_name : string
        The name of the parameter of interest that varies across the population.
    param_count : int
        The number of different values the parameter of interest will take on.
    center : float
        A measure of centrality for the distribution of the parameter of interest.
    spread : float
        A measure of spread or diffusion for the distribution of the parameter of interest.
    dist_type : string
        The type of distribution to be used.  Can be "lognormal" or "uniform" (can expand).

    Returns
    -------
    diff : float
        Difference between simulated and target capital to income ratio for this economy.
    '''
    Economy(LorenzBool = False, ManyStatsBool = False) # Make sure we're not wasting time calculating stuff
    Economy.distributeParams(param_name,param_count,center,spread,dist_type) # Distribute parameters
    Economy.solve()
    diff = Economy.calcKYratioDifference()
    print('getKYratioDifference tried center = ' + str(center) + ' and got ' + str(diff))
    return diff


def findLorenzDistanceAtTargetKY(Economy,param_name,param_count,center_range,spread,dist_type):
    '''
    Finds the sum of squared distances between simulated and target Lorenz points in an economy when
    a given parameter has heterogeneity according to some distribution.  The class of distribution
    and a measure of spread are given as inputs, but the measure of centrality such that the capital
    to income ratio matches the target ratio must be found.

    Parameters
    ----------
    Economy : cstwMPCmarket
        An object representing the entire economy, containing the various AgentTypes as an attribute.
    param_name : string
        The name of the parameter of interest that varies across the population.
    param_count : int
        The number of different values the parameter of interest will take on.
    center_range : [float,float]
        Bounding values for a measure of centrality for the distribution of the parameter of interest.
    spread : float
        A measure of spread or diffusion for the distribution of the parameter of interest.
    dist_type : string
        The type of distribution to be used.  Can be "lognormal" or "uniform" (can expand).

    Returns
    -------
    dist : float
        Sum of squared distances between simulated and target Lorenz points for this economy (sqrt).
    '''
    # Define the function to search for the correct value of center, then find its zero
    intermediateObjective = lambda center : getKYratioDifference(Economy = Economy,
                                                                 param_name = param_name,
                                                                 param_count = param_count,
                                                                 center = center,
                                                                 spread = spread,
                                                                 dist_type = dist_type)
    optimal_center = brentq(intermediateObjective,center_range[0],center_range[1],xtol=10**(-6))
    Economy.center_save = optimal_center

    # Get the sum of squared Lorenz distances given the correct distribution of the parameter
    Economy(LorenzBool = True) # Make sure we actually calculate simulated Lorenz points
    Economy.distributeParams(param_name,param_count,optimal_center,spread,dist_type) # Distribute parameters
    Economy.solveAgents()
    Economy.makeHistory()
    dist = Economy.calcLorenzDistance()
    Economy(LorenzBool = False)
    print ('findLorenzDistanceAtTargetKY tried spread = ' + str(spread) + ' and got ' + str(dist))
    return dist

def calcStationaryAgeDstn(LivPrb,terminal_period):
    '''
    Calculates the steady state proportions of each age given survival probability sequence LivPrb.
    Assumes that agents who die are replaced by a newborn agent with t_age=0.

    Parameters
    ----------
    LivPrb : [float]
        Sequence of survival probabilities in ordinary chronological order.  Has length T_cycle.
    terminal_period : bool
        Indicator for whether a terminal period follows the last period in the cycle (with LivPrb=0).

    Returns
    -------
    AgeDstn : np.array
        Stationary distribution of age.  Stochastic vector with frequencies of each age.
    '''
    T = len(LivPrb)
    if terminal_period:
        MrkvArray = np.zeros((T+1,T+1))
        top = T
    else:
        MrkvArray = np.zeros((T,T))
        top = T-1

    for t in range(top):
        MrkvArray[t,0] = 1.0 - LivPrb[t]
        MrkvArray[t,t+1] = LivPrb[t]
    MrkvArray[t+1,0] = 1.0

    w, v = np.linalg.eig(np.transpose(MrkvArray))
    idx = (np.abs(w-1.0)).argmin()
    x = v[:,idx].astype(float)
    AgeDstn = (x/np.sum(x))
    return AgeDstn

####################################################################################################

if __name__ == '__main__':

    # Set targets for K/Y and the Lorenz curve based on the data
    if Params.do_liquid:
        lorenz_target = np.array([0.0, 0.004, 0.025,0.117])
        KY_target = 6.60
    else: # This is hacky until I can find the liquid wealth data and import it
        lorenz_target = getLorenzShares(Params.SCF_wealth,weights=Params.SCF_weights,percentiles=Params.percentiles_to_match)
        lorenz_long_data = np.hstack((np.array(0.0),getLorenzShares(Params.SCF_wealth,weights=Params.SCF_weights,percentiles=np.arange(0.01,1.0,0.01).tolist()),np.array(1.0)))
        #lorenz_target = np.array([-0.002, 0.01, 0.053,0.171])
        KY_target = 10.26

    # Make AgentTypes for estimation
    if Params.do_lifecycle:
        DropoutType = cstwMPCagent(**Params.init_dropout)
        DropoutType.AgeDstn = calcStationaryAgeDstn(DropoutType.LivPrb,True)
        HighschoolType = deepcopy(DropoutType)
        HighschoolType(**Params.adj_highschool)
        HighschoolType.AgeDstn = calcStationaryAgeDstn(HighschoolType.LivPrb,True)
        CollegeType = deepcopy(DropoutType)
        CollegeType(**Params.adj_college)
        CollegeType.AgeDstn = calcStationaryAgeDstn(CollegeType.LivPrb,True)
        DropoutType.update()
        HighschoolType.update()
        CollegeType.update()
        EstimationAgentList = []
        for n in range(Params.pref_type_count):
            EstimationAgentList.append(deepcopy(DropoutType))
            EstimationAgentList.append(deepcopy(HighschoolType))
            EstimationAgentList.append(deepcopy(CollegeType))
    else:
        if Params.do_agg_shocks:
            PerpetualYouthType = cstwMPCagent(**Params.init_agg_shocks)
        else:
            PerpetualYouthType = cstwMPCagent(**Params.init_infinite)
        PerpetualYouthType.AgeDstn = np.array(1.0)
        EstimationAgentList = []
        for n in range(Params.pref_type_count):
            EstimationAgentList.append(deepcopy(PerpetualYouthType))

    # Give all the AgentTypes different seeds
    for j in range(len(EstimationAgentList)):
        EstimationAgentList[j].seed = j

    # Make an economy for the consumers to live in
    EstimationEconomy = cstwMPCmarket(**Params.init_market)
    EstimationEconomy.agents = EstimationAgentList
    EstimationEconomy.KYratioTarget = KY_target
    EstimationEconomy.LorenzTarget = lorenz_target
    EstimationEconomy.LorenzData = lorenz_long_data
    if Params.do_lifecycle:
        EstimationEconomy.PopGroFac = Params.PopGroFac
        EstimationEconomy.TypeWeight = Params.TypeWeight_lifecycle
        EstimationEconomy.T_retire = Params.working_T-1
        EstimationEconomy.act_T = Params.T_sim_LC
        EstimationEconomy.ignore_periods = Params.ignore_periods_LC
    else:
        EstimationEconomy.PopGroFac = 1.0
        EstimationEconomy.TypeWeight = [1.0]
        EstimationEconomy.act_T = Params.T_sim_PY
        EstimationEconomy.ignore_periods = Params.ignore_periods_PY
    if Params.do_agg_shocks:
        EstimationEconomy(**Params.aggregate_params)
        EstimationEconomy.update()
        EstimationEconomy.makeAggShkHist()

    # Estimate the model as requested
    if Params.run_estimation:
        # Choose the bounding region for the parameter search
        if Params.param_name == 'CRRA':
            param_range = [0.2,70.0]
            spread_range = [0.00001,1.0]
        elif Params.param_name == 'DiscFac':
            param_range = [0.95,0.995]
            spread_range = [0.006,0.008]
        else:
            print('Parameter range for ' + Params.param_name + ' has not been defined!')

        if Params.do_param_dist:
            # Run the param-dist estimation
            paramDistObjective = lambda spread : findLorenzDistanceAtTargetKY(
                                                            Economy = EstimationEconomy,
                                                            param_name = Params.param_name,
                                                            param_count = Params.pref_type_count,
                                                            center_range = param_range,
                                                            spread = spread,
                                                            dist_type = Params.dist_type)
            t_start = clock()
            spread_estimate = golden(paramDistObjective,brack=spread_range,tol=1e-4)
            center_estimate = EstimationEconomy.center_save
            t_end = clock()
        else:
            # Run the param-point estimation only
            paramPointObjective = lambda center : getKYratioDifference(Economy = EstimationEconomy,
                                                 param_name = Params.param_name,
                                                 param_count = Params.pref_type_count,
                                                 center = center,
                                                 spread = 0.0,
                                                 dist_type = Params.dist_type)
            t_start = clock()
            center_estimate = brentq(paramPointObjective,param_range[0],param_range[1],xtol=1e-6)
            spread_estimate = 0.0
            t_end = clock()

        # Display statistics about the estimated model
        #center_estimate = 0.986609223266
        #spread_estimate = 0.00853886395698
        EstimationEconomy.LorenzBool = True
        EstimationEconomy.ManyStatsBool = True
        EstimationEconomy.distributeParams(Params.param_name,Params.pref_type_count,center_estimate,spread_estimate,Params.dist_type)
        EstimationEconomy.solve()
        EstimationEconomy.calcLorenzDistance()
        print('Estimate is center=' + str(center_estimate) + ', spread=' + str(spread_estimate) + ', took ' + str(t_end-t_start) + ' seconds.')
        EstimationEconomy.center_estimate = center_estimate
        EstimationEconomy.spread_estimate = spread_estimate
        EstimationEconomy.showManyStats(Params.spec_name)

